/*
 * Copyright: (c) 2004-2007 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package edu.mayo.informatics.lexgrid.convert.directConversions;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.LexGrid.messaging.LgMessageDirectorIF;
import org.LexGrid.util.sql.DBUtility;
import org.apache.commons.collections.map.LRUMap;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import edu.mayo.informatics.lexgrid.convert.directConversions.UmlsCommon.UMLSBaseCode;

/**
 * A tool to load the LexGrid SQL table from UMLS - using the CUI's as the concept codes.
 * 
 * @author <A HREF="mailto:armbrust.daniel@mayo.edu">Dan Armbrust </A>
 * @author <A HREF="mailto:johnson.thomas@mayo.edu">Thomas Johnson</A>
 * @version subversion $Revision: 5996 $ checked in on $Date: 2007-09-28 03:15:27 -0500 (Fri, 28 Sep 2007) $
 */
//TODO need to set some associations as navigable, etc.
public class MetaThesaurusToSQL extends UMLSBaseCode
{
    int                       presentationCounter_ = 1;
    int                       codeCounter_         = 1;
    int                       definitionCounter_   = 1;
    int                       cuiCounter_          = 1;
    int                       semTypeCounter_      = 1;
    
    private String codingSchemeName_ = "NCI MetaThesaurus";
	private String relationsName_ = "Relations";

	// maintain load state ...
    private LRUMap  alreadyLoadedAssociations = new LRUMap(1);
    private String  lastAssociation;
    private HashSet loadedRelations_    = new HashSet();

    /**
     * Class to convert from UMLS to SQL.
     * 
     * @param sqlServer location of SQL server
     * @param sqlDriver SQL driver class
     * @param sqlUserName user name for server authentication
     * @param sqlPassword password for server authentication
     * @param umlsSezrver location of UMLS server
     * @param umlsDriver UMLS driver class
     * @param umlsUserName user name for server authentication
     * @param umlsPassword password for server authentication
     * @param codingScheme CodingScheme to be converted
     * @param enforceIntegrity enforce foreign key constraints on tables
     * @param director message director for log output
     * @throws Exception
     */
    public MetaThesaurusToSQL(String sqlServer, String sqlDriver, String sqlUserName, String sqlPassword, String tablePrefix,
            String umlsServer, String umlsDriver, String umlsUserName, String umlsPassword,
            boolean enforceIntegrity, LgMessageDirectorIF director) throws Exception
    {
        log = Logger.getLogger("convert.MetaThesaurusToSQL");
        messages_ = director;
        try
        {
            initIsoMap();
            //initSpecialCaseRelationPairs();
            initMRCONSOTTYMAP();

            makeConnections(umlsServer, umlsDriver, umlsUserName, umlsPassword, sqlServer, sqlDriver, sqlUserName,
                            sqlPassword, tablePrefix, enforceIntegrity);

            loadCodingScheme();
            closeConnections();
        }
        catch (Exception e)
        {
            log.error("Failed...", e);
            messages_.fatalAndThrowException("Failed...", e);
        }
        finally
        {
            if (sqlConnection_ != null)
            {
                sqlConnection_.close();
            }
            if (umlsConnection_ != null)
            {
                umlsConnection_.close();
            }
            if (umlsConnection2_ != null)
            {
                umlsConnection2_.close();
            }
        }
    }

    /*
     * kick off the process of loading the coding scheme
     */
    private void loadCodingScheme() throws SQLException
    {
        messages_.info("loading " + codingSchemeName_);

        String codingSchemeName = codingSchemeName_;
        String formalName = "NCI MetaThesaurus";
        String registeredName = getISOString(codingSchemeName);
        String defaultLanguage = "ENG";
        defaultLanguage_ = defaultLanguage;
        String representsVersion = getSourceVersionString("Unknown version of the MetaThesaurus");
        boolean isNative = true;
        int approxNumConcepts = 0; 
        String entityDescription = "NCI MetaThesaurus loaded from RRF files.";
        //TODO - this should be read from somewhere
        String copyright = "Copyright: (c) 2004-2007 Mayo Foundation for Medical Education and Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the triple-shield Mayo logo are trademarks and service marks of MFMER.Except as contained in the copyright notice above, or as used to identify MFMER as the author of this software, the trade names, trademarks, service marks, or product names of the copyright holder shall not be used in advertising, promotion or otherwise in connection with this software without prior written authorization of the copyright holder. Licensed under the Eclipse Public License, Version 1.0 (the \"License\"); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.eclipse.org/legal/epl-v10.html."; 

        messages_.info("Cleaning tables");
        sqlTableUtility_.cleanTables(codingSchemeName);

        int i = 1;
        insertIntoCodingScheme.setString(i++, codingSchemeName);
        insertIntoCodingScheme.setString(i++, formalName);
        insertIntoCodingScheme.setString(i++, registeredName);
        insertIntoCodingScheme.setString(i++, defaultLanguage_);
        insertIntoCodingScheme.setString(i++, representsVersion);
        DBUtility.setBooleanOnPreparedStatment(insertIntoCodingScheme, i++, new Boolean(isNative));
        insertIntoCodingScheme.setInt(i++, approxNumConcepts);
        if (sqlTableUtility_.getSQLTableConstants().supports2006Model())
        {
            DBUtility.setBooleanOnPreparedStatment(insertIntoCodingScheme, i++, null);
            DBUtility.setBooleanOnPreparedStatment(insertIntoCodingScheme, i++, null);
            DBUtility.setBooleanOnPreparedStatment(insertIntoCodingScheme, i++, null);
        }
        insertIntoCodingScheme.setString(i++, entityDescription);
        insertIntoCodingScheme.setString(i++, copyright);
        
        insertIntoCodingScheme.executeUpdate();

        messages_.info("Loaded 1 coding scheme");

        loadCodingSchemeMultiAttributes(codingSchemeName);
        loadCodingSchemeSupportedAttributes(codingSchemeName);

        loadConcepts(codingSchemeName);

        loadRelations(codingSchemeName);

        messages_.info("Adding the root relation node ...");

        // make sure that the required association parts exist
        if (!loadedRelations_.contains(codingSchemeName + relationsName_))
        {
            messages_.info("Adding the relation to put all of the associations under");

            addRelationToRelations(codingSchemeName, relationsName_, new Boolean(isNative), "parent child relationships");
        }

        buildRootNodes(codingSchemeName);
    }

    /*
     * load the codingSchemeMultiAttributes table
     */
    private void loadCodingSchemeMultiAttributes(String codingSchemeName) throws SQLException
    {
        messages_.info("loading coding scheme multi attributes");
        // things like localName - source (which has a locatorInfo)

        insertIntoCodingSchemeMultiAttributes.setString(1, codingSchemeName);
        insertIntoCodingSchemeMultiAttributes.setString(2, "source");
        insertIntoCodingSchemeMultiAttributes.setString(3, "RRF Files");
        if (stc_.supports2006Model())
        {
            insertIntoCodingSchemeMultiAttributes.setString(4, "");
            insertIntoCodingSchemeMultiAttributes.setString(5, "");
        }

        insertIntoCodingSchemeMultiAttributes.executeUpdate();

        insertIntoCodingSchemeMultiAttributes.setString(1, codingSchemeName);
        insertIntoCodingSchemeMultiAttributes.setString(2, "localName");
        insertIntoCodingSchemeMultiAttributes.setString(3, codingSchemeName);
        if (stc_.supports2006Model())
        {
            insertIntoCodingSchemeMultiAttributes.setString(4, "");
            insertIntoCodingSchemeMultiAttributes.setString(5, "");
        }

        insertIntoCodingSchemeMultiAttributes.executeUpdate();

        String iso = getISOString(codingSchemeName);
        insertIntoCodingSchemeMultiAttributes.setString(1, codingSchemeName);
        insertIntoCodingSchemeMultiAttributes.setString(2, "localName");
        insertIntoCodingSchemeMultiAttributes.setString(3, iso);
        if (stc_.supports2006Model())
        {
            insertIntoCodingSchemeMultiAttributes.setString(4, "");
            insertIntoCodingSchemeMultiAttributes.setString(5, "");
        }

        insertIntoCodingSchemeMultiAttributes.executeUpdate();
        
        insertIntoCodingSchemeMultiAttributes.setString(1, codingSchemeName);
        insertIntoCodingSchemeMultiAttributes.setString(2, "localName");
        insertIntoCodingSchemeMultiAttributes.setString(3, "10001");
        if (stc_.supports2006Model())
        {
            insertIntoCodingSchemeMultiAttributes.setString(4, "");
            insertIntoCodingSchemeMultiAttributes.setString(5, "");
        }

        insertIntoCodingSchemeMultiAttributes.executeUpdate();
        
        insertIntoCodingSchemeMultiAttributes.setString(1, codingSchemeName);
        insertIntoCodingSchemeMultiAttributes.setString(2, "localName");
        insertIntoCodingSchemeMultiAttributes.setString(3, "NCI_MetaThesaurus");
        if (stc_.supports2006Model())
        {
            insertIntoCodingSchemeMultiAttributes.setString(4, "");
            insertIntoCodingSchemeMultiAttributes.setString(5, "");
        }

        insertIntoCodingSchemeMultiAttributes.executeUpdate();
        
        if (iso.startsWith("urn:iso:"))
        {

            insertIntoCodingSchemeMultiAttributes.setString(1, codingSchemeName);
            insertIntoCodingSchemeMultiAttributes.setString(2, "localName");
            insertIntoCodingSchemeMultiAttributes.setString(3, iso.substring("urn:iso:".length()));
            if (stc_.supports2006Model())
            {
                insertIntoCodingSchemeMultiAttributes.setString(4, "");
                insertIntoCodingSchemeMultiAttributes.setString(5, "");
            }

            insertIntoCodingSchemeMultiAttributes.executeUpdate();
        }
    }

    /*
     * load the codingSchemeSupported Attributes table
     */
    private void loadCodingSchemeSupportedAttributes(String codingSchemeName) throws SQLException
    {
        messages_.info("loading supported attributes");

        // language - done later

        // format
        // arbitrary choices
        insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "Format", "text/plain", null, null, null, null);
        insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "Format", "text/html", null, null, null, null);
        insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "Format", "text/xml", null, null, null, null);

        // property - these are loaded after the concepts.

        // populate the supported coding schemes, now that we know what they are.
        String urn = getISOString(codingSchemeName);
        insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "CodingScheme", codingSchemeName, urn, null, null, null);

        messages_.info("Loaded 1 supported coding schemes");

        // Source
        // later

        // association
        populateSupportedAssociations();
        Set loadedNames = new HashSet();
        for (int i = 0; i < supportedAssociations_.length; i++)
        {
        	// Note: Registered association names can be duplicated
        	// in meta if defined for more than one source ontology.
        	String name = supportedAssociations_[i].name;
        	if (!loadedNames.contains(name))
        	{
        		insertIntoCodingSchemeSupportedAttributes(
			                                          codingSchemeName,
			                                          "Association",
			                                          (supportedAssociations_[i]).name,
			                                          null, null, null, null);
        		loadedNames.add(name);
        	}
        }
        messages_.info("added " + supportedAssociations_.length + " associations");

        // representationalForm - these are for tty codes in mrconso
        int i = 0;
        Enumeration values = mrconsoRepresentationalMap_.elements();
        while (values.hasMoreElements())
        {
            String var = (String) values.nextElement();
            insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "RepresentationalForm", var, null, null, null, null);
            i++;
        }
        messages_.info("Loaded " + i + " representational forms");
    }
    
    /*
     * load the concepts for a coding scheme
     */
    private void loadConcepts(String codingSchemeName) throws SQLException
    {
        messages_.info("loading concepts - getting a total count");

        PreparedStatement getCodingSchemeInfo = umlsConnection2_
                .prepareStatement("SELECT  COUNT(*) as cnt FROM MRCONSO");

        ResultSet results = getCodingSchemeInfo.executeQuery();

        results.next();
        int total = results.getInt("cnt");

        results.close();
        getCodingSchemeInfo.close();

        int start = 0;
        String lastCui = null;
        ArrayList conceptPresentations = new ArrayList();
        int codeCount = 0;
        int rowCount = 0;

        // collect them all into batches of concept codes. when the code changes, load that code
        while (start < total)
        {
            // mysql still can't seem to stream results properly, so I'm going to batch them.
            messages_.info("Fetching a batch of results");
            getCodingSchemeInfo = umlsConnection2_
                    .prepareStatement(umlsSqlModifier_.modifySQL("SELECT  LAT, CODE, CUI, TS, STT, ISPREF, STR, TTY, SAB FROM MRCONSO ORDER BY {BINARY} CUI {LIMIT}"));

            //mysql doesn't stream results - the {LIMIT above and this is for getting limits on mysql code}
            if (umlsSqlModifier_.getDatabaseType().equals("MySQL"))
            {
                getCodingSchemeInfo.setInt(1, start);
                getCodingSchemeInfo.setInt(2, batchSize);
                start += batchSize;
            }
            else if (umlsSqlModifier_.getDatabaseType().equals("PostgreSQL"))
            {
                //postgres properly streams results, we can just set the fetch size, and only loop once
                getCodingSchemeInfo.setFetchSize(batchSize);
                umlsConnection2_.setAutoCommit(false);
                start = total;
            }
            else
            {
                start = total;
            }

            results = getCodingSchemeInfo.executeQuery();

            log.debug("query finished, processing results");

            while (results.next())
            {
                rowCount++;
                // store all the data from this row.

                ConceptPresentation temp = new ConceptPresentation();
                temp.conceptCode = results.getString("CODE");
                temp.cui = results.getString("CUI");

                if (lastCui != null && !temp.cui.equals(lastCui))
                {
                    loadConcept(codingSchemeName, (ConceptPresentation[]) conceptPresentations
                            .toArray(new ConceptPresentation[conceptPresentations.size()]));

                    conceptPresentations.clear();
                    presentationCounter_ = 1;
                    codeCounter_ = 1;
                    definitionCounter_ = 1;
                    cuiCounter_ = 1;
                    semTypeCounter_ = 1;

                    if (codeCount % 100 == 0)
                    {
                        messages_.busy();
                    }

                    codeCount++;
                    if (codeCount % 10000 == 0)
                    {
                        messages_.info("On row "
                                + rowCount
                                + " out of "
                                + total
                                + " rows - found "
                                + codeCount
                                + " concepts");
                    }
                }

                temp.language = results.getString("LAT");

                // add for populating the supportedLanguages later
                supportedLanguages_.add(temp.language);

                temp.presentationFormat = null;
                temp.TTY = results.getString("TTY");
                temp.representationForm = temp.TTY;
                temp.value = results.getString("STR");
                temp.TS = results.getString("TS");
                temp.STT = results.getString("STT");
                temp.isPreferred = new Boolean(false);
                temp.source = results.getString("SAB");
                temp.ISPREF = results.getString("ISPREF");

                // See if there is a "better" value for the representationalForm mapping
                String repFormMap = (String) mrconsoRepresentationalMap_.get(temp.representationForm);
                if (repFormMap != null)
                {
                    temp.representationForm = repFormMap;
                }

                lastCui = temp.cui;
                conceptPresentations.add(temp);
            }
            results.close();
            getCodingSchemeInfo.close();
        }

        // load the last one
        if (lastCui != null)
        {
            // need to add this, just once - assuming there were codes, so this is a good place for it.
            supportedPropertyTypes_.add("textualPresentation");
            loadConcept(codingSchemeName, (ConceptPresentation[]) conceptPresentations
                    .toArray(new ConceptPresentation[conceptPresentations.size()]));

            codeCount++;
        }

        messages_.info("Loaded " + codeCount + " concepts from  " + rowCount + " rows");
        
        supportedPropertyQualifiers_.add("source-code");
        
        updateApproxNumberOfConcepts(codeCount, codingSchemeName);
        loadSupportedProperties(codingSchemeName);
        loadSupportedPropertyQualifiers(codingSchemeName);
        loadSupportedLanguages(codingSchemeName);
        loadSupportedSources(codingSchemeName);
    }
    
    /*
     * Helper method to load all the properties and presentations of a concept
     */
    private void loadConcept(String codingSchemeName, ConceptPresentation[] conceptPresentations) throws SQLException
    {
        // order them - this will organize them by language, and put the "best" presentation
        // first for each language. default language will be the at the very top.
        Arrays.sort(conceptPresentations, new ConceptPresentationSorter(true));

        // now set each isPreferred to true for the first entry of each language
        String prevLang = conceptPresentations[0].language;
        conceptPresentations[0].isPreferred = new Boolean(true);
        for (int i = 1; i < conceptPresentations.length; i++)
        {
            if (!conceptPresentations[i].language.equals(prevLang))
            {
                conceptPresentations[i].isPreferred = new Boolean(true);
            }
            prevLang = conceptPresentations[i].language;
        }

        // load the entry into the concepts table - must be at least 1 row. the first one will
        // be the best text presentation to use here.
        try
        {
            addConceptToConcepts(codingSchemeName, conceptPresentations[0].cui, null, null, null, new Boolean(true),
                                 "Active", new Boolean(false), conceptPresentations[0].value);
        }
        catch (SQLException e)
        {
            // Recover from a failure adding a new code. Likely caused by trying
            // to insert a duplicate code - some code systems are case sensitive - and
            // most of our databases aren't.
            log.error("Problem inserting new code " + conceptPresentations[0].conceptCode, e);
            messages_.info("ERROR - Problem inserting new code " + conceptPresentations[0].conceptCode);
            return;
        }

        // load its properties
        for (int i = 0; i < conceptPresentations.length; i++)
        {
            // put in the presentation
            addToConceptProperty(codingSchemeName, conceptPresentations[i].cui, "T-" + presentationCounter_,
                                 "presentation", "textualPresentation", conceptPresentations[i].language,
                                 conceptPresentations[i].presentationFormat, conceptPresentations[i].isPreferred,
                                 null, null, conceptPresentations[i].representationForm, conceptPresentations[i].value);

            addConceptToConceptPropertyMultiAttributes(codingSchemeName, conceptPresentations[i].cui, "T-"
                    + presentationCounter_, "source", conceptPresentations[i].source, "", "");
            
            // add this source to the source list in use...
            supportedSources_.add(conceptPresentations[i].source);
            

            if (conceptPresentations[i].conceptCode != null && conceptPresentations[i].conceptCode.length() > 0)
            {
                addConceptToConceptPropertyMultiAttributes(codingSchemeName, conceptPresentations[i].cui, "T-"
                                                       + presentationCounter_, "qualifier", "source-code", conceptPresentations[i].conceptCode, "");
             
            }
            
            presentationCounter_++;
        }

        // load the definitions
        try
        {
            addToDefinitions(conceptPresentations[0].cui, codingSchemeName);
        }
        catch (SQLException e)
        {
            //recover from an error loading definitions
            log.error("Problem loading definitions for code " + conceptPresentations[0].cui, e);
            messages_.info("ERROR - Problem inserting new code " + conceptPresentations[0].cui);

        }

        // load the other properties
        loadOtherProperties(conceptPresentations[0].cui, codingSchemeName);

    }

	/*
     * populate the properties attributes of the codingSchemeSuppportedAttributes table
     */
    private void loadSupportedCodingSchemes(String codingSchemeName) throws SQLException
    {
        messages_.info("Adding the Supported Coding Schemes.");
        Iterator iter = supportedCodingSchemes_.iterator();
        
        int count = 0;
        while (iter.hasNext())
        {
            String value = (String)iter.next();
            count++;
            // populate the supported coding schemes, now that we know what they are.
            String urn = getISOString(value);
            insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "CodingScheme", value, urn, null, null, null);
        }
        
        messages_.info("Added " + count + "  Supported Coding Schemes.");

    }
    
    /*
     * Load the definitions for a concept
     */
    private void addToDefinitions(String cui, String codingSchemeName) throws SQLException
    {
        PreparedStatement getDefinitions = umlsConnection_.prepareStatement("SELECT DEF, SAB FROM MRDEF WHERE CUI = ?");

        getDefinitions.setString(1, cui);

        ResultSet definitions = getDefinitions.executeQuery();
        boolean isPreferred = true;
        while (definitions.next())
        {
            supportedPropertyTypes_.add("definition");
            addToConceptProperty(codingSchemeName, cui, "D-" + definitionCounter_, "definition", "definition", null, null,
                                 new Boolean(isPreferred), null, null, null, definitions.getString("DEF"));
            isPreferred = false;
            // umls doesn't seem to have the notion of a preferred definition, so I'll just set the first one to
            // preferred.

            addConceptToConceptPropertyMultiAttributes(codingSchemeName, cui, "D-" + definitionCounter_++, "source",
                                                       definitions.getString("SAB"), "", "");
        }
        getDefinitions.close();
    }

    /*
     * Load other properties - just semantic types now.
     */
    private void loadOtherProperties(String cui, String codingSchemeName) throws SQLException
    {
        // get the semantic type(s)
        supportedPropertyTypes_.add("Semantic_Type");
        PreparedStatement getSemTypes = umlsConnection_.prepareStatement("SELECT STY FROM MRSTY WHERE CUI = ?");
        getSemTypes.setString(1, cui);

        ResultSet semTypes = getSemTypes.executeQuery();
        while (semTypes.next())
        {
            addToConceptProperty(codingSchemeName, cui, "SemType-" + semTypeCounter_++, "property", "Semantic_Type", null, null,
                                 null, null, null, null, semTypes.getString("STY"));
        }
        getSemTypes.close();
    }
    
    public static void main(String[] args) throws Exception
    {
        //Logger.getRootLogger().setLevel(Level.OFF);
//        new MetaThesaurusToSQL("jdbc:postgresql://mir04/LexGrid2", "org.postgresql.Driver", "mirpub", "mirpub",
//                "jdbc:mysql://mir04/MetaThesaurus", "org.gjt.mm.mysql.Driver", "mirpub", "mirpub", "NCI MetaThesaurus",
//                true, new CommandLineMessageDirector());
    }

    /**
     * @return the codingSchemeName
     */
    public String getCodingSchemeName()
    {
        return this.codingSchemeName_;
    }

    ////////////////////////////////////////////////////////////////
    // Code for tracking associations used in each container, 
    // root nodes, and hierarchical navigation ...
    ////////////////////////////////////////////////////////////////

    /**
     * Inner class to track keyed references to source qualifiers.
     */
    protected class QualifierHolder {
    	String multiKey = null;
    	HashSet sources = new HashSet(); 
		protected QualifierHolder(String key) {
			multiKey = key;
		}
	}

    /**
     * Add a code to code association.
	 * @param currentCodingScheme
     * @param sourceCode
     * @param association
     * @param targetCode
     * @param sourceOfAssociation
     * @param forwardName
     * @param reverseName
     * @param count
     * @return true if the association was added; false if an error
     * occurred or the association was already defined.
     * @throws SQLException
     */
    protected boolean addConceptAssociationToConceptsHelper(String currentCodingScheme, String sourceCode,
            String association, String targetCode, String sourceOfAssociation, String forwardName, String reverseName,
            int count) throws SQLException
    {
        String relation = relationsName_;
        String relationDescription = "MetaThesaurus Relationships";

        // Register the higher level relations container if not present ...
        if (!loadedRelations_.contains(currentCodingScheme + relation)) {
            messages_.info("Adding the relation to put all of the associations under.");
            addRelationToRelations(currentCodingScheme, relation, Boolean.TRUE, relationDescription);
            loadedRelations_.add(currentCodingScheme + relation);
        }

        // Register the higher level association if not present ...
        if (!isLoaded(association, currentCodingScheme, relation)) {
            messages_.info("Adding the association to put sources and targets under.");
            boolean navigable = true;
            Boolean transitive = new Boolean(
            	isHierarchicalName(forwardName, currentCodingScheme) ||
            	isHierarchicalName(reverseName, currentCodingScheme));
            addAssociationToAssociations(
            	currentCodingScheme, relation, association,
            	forwardName, reverseName, "",
				new Boolean(navigable), transitive, null, null, null, null,
				null, null, null, null, null,
				"MetaThesaurus Associations");
            markLoaded(association, currentCodingScheme, relation);
        }

        // Finally, register the instance of the association (concept source to target) ...
        // Return value - default indicates no association was loaded.
        boolean isAdded = false;
        boolean isDuplicate = false;

        // The cache key for this association.
        // Note: no need for source or target coding scheme names in here - always NCI MetaThesaurus.
        // Shouldn't need the association name either, since we go through them in a known order (per SAB).
        String assocKey = sourceCode + ":" + targetCode;
        
        // The key used to map between association and association qualifiers.
        String multiKey = null;
        QualifierHolder qHolder = (QualifierHolder) alreadyLoadedAssociations.get(assocKey);

        // Check the cache... already loaded this one?  If so, move on to qualification.
        if (qHolder != null)
        	multiKey = qHolder.multiKey;
        else {
        	// Note:  LRU cache helps avoid duplicates, but is not a guarantee
        	// given current order of processing of items and associations per
        	// source (if same source/target/assoc combo exists for two
        	// separate SABs, the cache entry may be dropped between references).

        	// Try optimistic insert.  On error, check for duplicate and reuse
        	// multiAttributesKey if present.  If not, throw the original err.
        	try {
				multiKey =
					generateUniqueKey(new String[] {
						currentCodingScheme, currentCodingScheme, 
						sourceCode, relation, association, 
						currentCodingScheme, targetCode
						});
		        addConceptAssociationToConcepts(
		        	currentCodingScheme,
		        	currentCodingScheme, sourceCode,
		        	relation, association,
		        	currentCodingScheme, targetCode,
		        	multiKey, null, null);		    			        
        		isAdded = true;
        	} catch (SQLException sqle) {
	        	// Check for existing association (same link may be defined
				// for multiple sources in meta, etc. If found, extract the
        		// existing multi-attribute key for reuse to add qualifiers.
				ResultSet rs = null;
				getAssocInstance_.setString(1, currentCodingScheme);
				getAssocInstance_.setString(2, relation);
				getAssocInstance_.setString(3, association);
				getAssocInstance_.setString(4, currentCodingScheme);
				getAssocInstance_.setString(5, sourceCode);
				getAssocInstance_.setString(6, currentCodingScheme);
				getAssocInstance_.setString(7, targetCode);
				multiKey = null;
				try {
					rs = getAssocInstance_.executeQuery();
					if (rs.next()) {
						multiKey = rs.getString("multiAttributesKey");
						isDuplicate = true;
					} else {
						throw sqle;
					}
				} catch (Exception e) {
        			log.warn("Association instance not inserted.", sqle);
				} finally {
					if (rs != null)
						rs.close();
				}
			}
        }
        // Ensure registration of source qualifier, and cache for
        // future reference.  Here we have no alternate path to follow
        // if an error occurs and the qualifier does or does not exist.
        // The load will not be terminated on failure to load a
        // qualifier.  On error, log a warning and continue.
        if ((isAdded || isDuplicate)
        		&& multiKey != null
        		&& StringUtils.isNotBlank(sourceOfAssociation))
        {
        	if (qHolder == null) {
        		qHolder = new QualifierHolder(multiKey);
        		alreadyLoadedAssociations.put(assocKey, qHolder);
        	}
        	if (!qHolder.sources.contains(sourceOfAssociation)) {
        		qHolder.sources.add(sourceOfAssociation);
        		try {
        			addConceptAssociationQualifierToConceptAssociation(
        				currentCodingScheme, multiKey, sourceOfAssociation, "Source");
        		} catch (SQLException sqle) {
					ResultSet rs2 = null;
					getAssocQualifier_.setString(1, currentCodingScheme);
					getAssocQualifier_.setString(2, multiKey);
					getAssocQualifier_.setString(3, sourceOfAssociation);
					getAssocQualifier_.setString(4, "Source");
					try {
						rs2 = getAssocQualifier_.executeQuery();
						if (!rs2.next())
							throw sqle;
					} catch (Exception e) {
	        			log.warn("Association qualifier not inserted.", sqle);
					} finally {
						if (rs2 != null)
							rs2.close();
					}
        		}
        	}
        }
        return isAdded;
    }
    
	/**
	 * Build navigable relations from special endpoint nodes '@' and '@@' to
	 * first level nodes in hierarchical associations.
	 * 
	 * @param codingSchemeName
	 *            The LexGrid code system name.
	 */
    protected void buildRootNodes(String codingSchemeName) throws SQLException {
    	// Calculate for the source relations container ...
    	SABString[] names = getHierRelas();
    	for (int i = 0; i < names.length; i++) {
    		SABString name = names[i];
    		if (isLoaded(name.str, codingSchemeName, relationsName_)) {
            	buildRootNode(codingSchemeName, true, new SABString[] {name}, relationsName_);
    		}
    	}
    	
    	// Calculate for standard umls relations ...
    	names = getHierRels();
    	for (int i = 0; i < names.length; i++) {
    		SABString name = names[i];
    		if (isLoaded(name.str, codingSchemeName, relationsName_)) {
            	buildRootNode(codingSchemeName, true, new SABString[] {name}, relationsName_);
    		}
    	}
    }
    
    /**
     * High level method with responsibility to map relationship content
     * from UMLS RRF-based format to LexGrid model.
     * @param codingSchemeName
     * @throws SQLException
     */
    protected void loadRelations(String codingSchemeName) throws SQLException {
		// Sort the associations, loading RELA's first and keeping
		// non-directional names (potentially qualified by different SABs)
		// together.
		Arrays.sort(supportedAssociations_, new AssociationSorter());
        
		for (int i = 0; i < supportedAssociations_.length; i++) {
			Association assoc = (Association) supportedAssociations_[i];
			if (lastAssociation == null || !lastAssociation.equals(assoc.name)) {
				// Moving on to a new association name.
				lastAssociation = assoc.name;
				// Reset the map used to track source/target combinations loaded.
				// 100,000 should be more than big enough for the other associations,
				// items are ordered by source code.
				alreadyLoadedAssociations.clear();          
				alreadyLoadedAssociations = new LRUMap(100000);
			}

			int count = loadRelationsHelper(assoc, codingSchemeName);
			if (count == 0)
			{
				messages_.info("No relations were found for " + assoc.toShortString());
				log.warn("No relations were found for " + assoc.toShortString());
			}
		}

		// Now we know all the supported coding schemes (some may have been
		// identified as source or target while processing relationships)
		// and association qualifiers.  Add them now to the coding scheme
		// metadata.
		loadSupportedCodingSchemes(codingSchemeName);
		loadSupportedAssociationQualifiers(codingSchemeName);
    }

    /**
     * Carry out the mapping of relationships from UMLS RRF-based format to
     * LexGrid model.
     * @param assoc
     * @param codingSchemeName
     * @return The number of association instances added to the LexGrid repository.
     * @throws SQLException
     */
    protected int loadRelationsHelper(Association assoc, String codingSchemeName)
			throws SQLException
	{
    	// Statement to get count of items to process,
    	// qualified by SAB if association is SAB-specific ...
    	StringBuffer sb;
    	sb = new StringBuffer(256)
    		.append("SELECT COUNT(*) AS cnt FROM MRREL WHERE ")
    		.append(assoc.rrfField).append(" = ?");
    	if (StringUtils.isNotBlank(assoc.rrfSAB))
    		sb.append(" AND SAB = ?");
    	PreparedStatement getRelationsCount = umlsConnection2_.prepareStatement(
    		sb.toString());

    	// Statement to get items to process,
    	// qualified by SAB if association is SAB-specific ...
		// Note: ordered to enable a cache to not load duplicates.
    	sb = new StringBuffer(256)
			.append("SELECT CUI1, CUI2, SAB FROM MRREL WHERE ")
			.append(assoc.rrfField).append(" = ?");
    	if (StringUtils.isNotBlank(assoc.rrfSAB))
    		sb.append(" AND SAB = ?");
    	sb.append(" ORDER BY CUI1 {LIMIT}");
    	PreparedStatement getRelations = umlsConnection2_.prepareStatement(
    		umlsSqlModifier_.modifySQL(sb.toString()));

       	int count = 0;
    	try {
        	// Retrieve and print the count for progress monitoring ... 
    		int start = 0;
    		int total = 0;
    		getRelationsCount.setString(1, assoc.rrfName);
    		if (StringUtils.isNotBlank(assoc.rrfSAB))
        		getRelationsCount.setString(2, assoc.rrfSAB);

    		messages_.info("Counting relations for " + assoc.rrfName);
    		ResultSet results = getRelationsCount.executeQuery();
    		results.next();
    		total = results.getInt("cnt");
    		results.close();

    		// Process matching entries ...
    		while (start < total) {
    			int param = 0;
    			getRelations.setString(++param, assoc.rrfName);
        		if (StringUtils.isNotBlank(assoc.rrfSAB))
            		getRelations.setString(++param, assoc.rrfSAB);

    			// MySQL doesn't stream results - the {LIMIT above and this is for getting limits on MySQL code}
    			if (umlsSqlModifier_.getDatabaseType().equals("MySQL")) {
    				getRelations.setInt(++param, start);
    				getRelations.setInt(++param, batchSize * 3);
    				start += batchSize * 3;
    			}
				// PostgreSQL properly streams results, we can just set the fetch size, and only loop once
    			else if (umlsSqlModifier_.getDatabaseType().equals("PostgreSQL")) {
    				getRelations.setFetchSize(batchSize * 3);
    				umlsConnection2_.setAutoCommit(false);
    				start = total;
    			}
    			else {
    				start = total;
    			}

    			messages_.info("Getting a batch of relations for " + assoc.rrfName);
    			results = getRelations.executeQuery();

    			while (results.next()) {
    				// If the source association definition is reversed, we need to flip
    				// the source and target concepts accordingly.
    				boolean isReversed = assoc.rrfSourceDirectionalityReversed;
    				String sourceCode = results.getString(isReversed ? "CUI2" : "CUI1");
    				String targetCode = results.getString(isReversed ? "CUI1" : "CUI2");
    				String sab = results.getString("SAB");

    				boolean result = addConceptAssociationToConceptsHelper(
    						codingSchemeName, sourceCode,
    						assoc.name, targetCode, sab,
    						assoc.rrfName, assoc.rrfInverse, count);
    				if (result)
    					count++;
    				if (count % 100 == 0)
    					messages_.busy();
    				if (count % 10000 == 0)
    					messages_.info("Loaded " + count + " out of a possible total of " + total);
    			}
    			results.close();
    		}
    		messages_.info("Loaded " + count + " out of a possible total of " + total);
    	}
    	finally {
    		getRelations.close();
    		getRelationsCount.close();
    	}
    	return count;
    }

    /**
     * Detect and register association information to the array of supported
     * associations maintained by the loader.
     * @throws SQLException
     */
    protected void populateSupportedAssociations() throws SQLException {
        messages_.info("Getting the descriptive associations");
        Map relationsHolder = new Hashtable();

        PreparedStatement getRelations = umlsConnection_.prepareStatement("SELECT DISTINCT RELA, SAB, DIR FROM MRREL");
        try {
        	ResultSet relations = getRelations.executeQuery();
	        while (relations.next()) {
	            String temp = relations.getString("RELA");
	            String tempDirFlag = relations.getString("DIR");
	            String tempSAB = relations.getString("SAB");
	            if (temp == null || temp.length() == 0)
	                continue;
	            mapSupportedAssociationsHelper(temp, tempSAB, tempDirFlag, "RELA", relationsHolder);
	        }
	        relations.close();
        } finally {
        	getRelations.close();
        }

        messages_.info("Getting the base associations");
        getRelations = umlsConnection_
                .prepareStatement("SELECT DISTINCT REL, SAB, DIR FROM MRREL");
        try {
        	ResultSet relations = getRelations.executeQuery();
	        while (relations.next()) {
	            String temp = relations.getString("REL");
	            String tempDirFlag = relations.getString("DIR");
	            String tempSAB = relations.getString("SAB");
	            if (temp.length() == 0)
	                continue;
	            mapSupportedAssociationsHelper(temp, tempSAB, tempDirFlag, "REL", relationsHolder);
	        }
	        relations.close();
        } finally {
        	getRelations.close();
        }

        supportedAssociations_ = (Association[])
        	relationsHolder.values().toArray(new Association[relationsHolder.size()]);
        messages_.info("done processing relations");
    }

}