/*
 * Copyright: (c) 2004-2007 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package edu.mayo.informatics.lexgrid.convert.directConversions;

import java.security.MessageDigest;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;

import org.LexGrid.messaging.LgMessageDirectorIF;
import org.LexGrid.util.sql.DBUtility;
import org.LexGrid.util.sql.lgTables.SQLTableConstants;
import org.apache.commons.collections.OrderedMapIterator;
import org.apache.commons.collections.map.LRUMap;
import org.apache.commons.collections.map.ListOrderedMap;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import edu.mayo.informatics.lexgrid.convert.directConversions.UmlsCommon.UMLSBaseCode;

/**
 * A tool to load the LexGrid SQL table from UMLS.
 * 
 * @author <A HREF="mailto:armbrust.daniel@mayo.edu">Dan Armbrust</A>
 * @author <A HREF="mailto:erdmann.jesse@mayo.edu">Jesse Erdmann</A>
 * @author <A HREF="mailto:johnson.thomas@mayo.edu">Thomas Johnson</A>
 * @version subversion $Revision: 5996 $ checked in on $Date: 2007-09-28 03:15:27 -0500 (Fri, 28 Sep 2007) $
 */
public class UMLSToSQL extends UMLSBaseCode
{
	// maintain load state ...
    private String loadedCodingSchemeName_ = "";
	private HashSet alreadyLoadedAssociations = new HashSet();
	private HashSet loadedRelations_ = new HashSet();
	private String lastAssociation;
	
    // the map and the static int's are used to map the mrsat entries to the proper places.
    // see the initMRSATMap method.
    private Hashtable            mrsatMap_                   = new Hashtable();
    private Hashtable            atnRepresentationalMap_     = new Hashtable();
    private LRUMap               cuiToCodeCache_             = new LRUMap(50000);
    private LRUMap               auiToCodeCache_             = new LRUMap(50000);
    private static final Integer COMMENT                     = new Integer(1);
    private static final Integer INSTRUCTION                 = new Integer(2);
    private static final Integer PRESENTATION                = new Integer(3);
    private static final Integer SKIP                        = new Integer(4);

    int                          instructionCounter_         = 1;
    int                          presentationCounter_        = 1;
    int                          commentCounter_             = 1;
    int                          definitionCounter_          = 1;
    int                          propertyCounter_            = 1;
    int                          cuiCounter_                 = 1;
    int                          semTypeCounter_             = 1;

    PreparedStatement getSnomedLanguage_;
    
	// Define separate containers for standard UMLS relations and
    // unique relation name attributes defined per source.
	private static final String SRC_RELATIONS = "Relations";
	private static final String STD_RELATIONS = "UMLS_Relations";

	// To control tagging of associations that are contextually linked
	// via individual text representations (AUIs) rather than at
	// the concept level.  We handle through qualification by HCD value
	// of the involved text properties and associations, so that
	// contextual chains are preserved while full-fledged relations
	// remain defined at the concept level.
	private int mrhierOpt_ = 1;

	// Supporting inner classes ...
    private class AUI_PROP {
		String atn = "";
		String atv = "";
		String stype = "";
		String metaui = "";
		
		AUI_PROP(String atn, String atv, String stype, String metaui) {
			this.atn = atn;
			this.atv = atv;
			this.stype = stype;
			this.metaui = metaui;
		}
	}

	private class RelationType
	{
	    String  relation;
	    String  relationDescription;
	    boolean isNative;
	
	    public RelationType(String relation, String relationDescription, boolean isNative)
	    {
	        this.relation = relation;
	        this.relationDescription = relationDescription;
	        this.isNative = isNative;
	    }
	
	}

	/**
    * Class to convert from UMLS to SQL.
    * @param sqlServer location of SQL server
    * @param sqlDriver SQL driver class
    * @param sqlUserName username for server authentication
    * @param sqlPassword password for server authentication
    * @param sqlTablePrefix table prefix to use for generated files
    * @param umlsServer location of UMLS server
    * @param umlsDriver UMLS driver class
    * @param umlsUserName username for server authentication
    * @param umlsPassword password for server authentication
    * @param codingScheme CodingScheme to be converted
    * @param enforceIntegrity enforce foreign key constraints on tables
    * @param mrHierOpt indicates level of processing for MRHIER entries
    * @param director message director for log output
    * @throws Exception
    */
    public UMLSToSQL(String sqlServer,
    		String sqlDriver,
    		String sqlUserName,
    		String sqlPassword,
    		String sqlTablePrefix,
    		String umlsServer,
            String umlsDriver,
            String umlsUserName,
            String umlsPassword,
            String codingScheme,
            boolean enforceIntegrity,
            int mrhierOpt,
            LgMessageDirectorIF director) throws Exception
    {
        log                         = Logger.getLogger("convert.UMLSToSQL");
        messages_ = director;
        try
        {
            initIsoMap();
            initMRSATMap();
            initMRCONSOTTYMAP();

            makeConnections(umlsServer, umlsDriver, umlsUserName, umlsPassword, sqlServer, sqlDriver, sqlUserName,
                            sqlPassword, sqlTablePrefix, enforceIntegrity);

            initOtherSqlStatements();

            // If they didn't pass a coding scheme, just create the tables and exit.
            if (codingScheme == null || codingScheme.trim().length() == 0)
            {
                return;
            }

            mrhierOpt_ = mrhierOpt;
            loadCodingScheme(codingScheme);
            closeOtherSqlStatement();
            closeConnections();
        }
        catch (Exception e)
        {
            log.error("Failed...", e);
            messages_.fatalAndThrowException("Failed", e);
        }
        finally
        {
            if (sqlConnection_ != null)
            {
                sqlConnection_.close();
            }
            if (umlsConnection_ != null)
            {
                umlsConnection_.close();
            }
            if (umlsConnection2_ != null)
            {
                umlsConnection2_.close();
            }
        }
    }

    public String getLoadedCodingSchemeName()
    {
        return loadedCodingSchemeName_;
    }
   
    /*
	 * enter the mappings from MRSAT ATN codes to the locations we want them.
	 */
	private void initMRSATMap()
	{
	    mrsatMap_.put("AN", COMMENT);
	    mrsatMap_.put("CX", COMMENT);
	    mrsatMap_.put("HN", COMMENT);
	
	    mrsatMap_.put("EV", PRESENTATION);
	    atnRepresentationalMap_.put("EV", "Abbrev");
	
	    mrsatMap_.put("DID", SKIP);
	    mrsatMap_.put("MUI", SKIP);
	
	    // SNOMEDCT puts the language codes in attributes, instead of putting them on the concept
	    // don't want to add them as attributes - these will be pulled out in a custom case in the load
	    // concept options as necessary.
	    mrsatMap_.put("LANGUAGECODE", SKIP);
	    mrsatMap_.put("SUBSETLANGUAGECODE", SKIP);
	
	    // also don't need these from snomed - because they don't really make sense the way
	    // that we load them - they have these for each presentation - while in our representation
	    // they attributes apply to the entire concept
	    mrsatMap_.put("DESCRIPTIONSTATUS", SKIP);
	    mrsatMap_.put("DESCRIPTIONTYPE", SKIP);
	    mrsatMap_.put("INITIALCAPITALSTATUS", SKIP);
	    mrsatMap_.put("CHARACTERISTICTYPE", SKIP);
	    mrsatMap_.put("REFINABILITY", SKIP);
	    mrsatMap_.put("SUBSETMEMBER", SKIP);
	}

	private void initOtherSqlStatements() throws SQLException
	{
	    getSnomedLanguage_ = umlsConnection_
	    	.prepareStatement("SELECT ATV FROM MRSAT WHERE CUI = ? AND SAB = 'SNOMEDCT' AND LUI = ? AND SUI = ? AND (ATN = 'LANGUAGECODE' OR ATN = 'SUBSETLANGUAGECODE')");
	}

	private void closeOtherSqlStatement() throws SQLException
	{
	    if (getSnomedLanguage_ != null)
	    {
	        getSnomedLanguage_.close();
	    }
	}

	/*
     * kick off the process of loading the coding scheme
     */
    private void loadCodingScheme(String UMLSCodingSchemeName) throws SQLException
    {
        messages_.info("Loading " + UMLSCodingSchemeName);
        PreparedStatement getCodingSchemeInfo = umlsConnection_
                .prepareStatement("SELECT  SON, SVER, SCC, LAT, SSN, SCIT, TFR FROM MRSAB WHERE RSAB = ?");
        getCodingSchemeInfo.setString(1, UMLSCodingSchemeName);

        ResultSet results = getCodingSchemeInfo.executeQuery();

        if (results.next())
        {
            String codingSchemeName = results.getString("SSN");
            loadedCodingSchemeName_ = codingSchemeName;
            String formalName = results.getString("SON");
            String registeredName = getISOString(UMLSCodingSchemeName);
            String defaultLanguage = results.getString("LAT");
            defaultLanguage_ = defaultLanguage;
            if (UMLSCodingSchemeName.equals("SNOMEDCT"))
            {
                //snomed has its languages in a weird way - hard code this.
                defaultLanguage_ = "en";
            }
            String representsVersion = results.getString("SVER");
            if (representsVersion == null || representsVersion.length() == 0)
            {
                messages_.warn("The Version number could not be read from the MRSAB file, SVER column.");
                representsVersion = "MISSING";
            }
            boolean isNative = false;
            int approxNumConcepts = 0;
            String anc = results.getString("TFR");
            try
            {
                approxNumConcepts = Integer.parseInt(anc);
            }
            catch (Exception e)
            {
                //do nothing
            }
            String entityDescription = results.getString("SCIT");
            String copyright = results.getString("SCC");

            messages_.info("Cleaning tables");
            sqlTableUtility_.cleanTables(codingSchemeName);

            int i = 1;
            insertIntoCodingScheme.setString(i++, codingSchemeName);
            insertIntoCodingScheme.setString(i++, formalName);
            insertIntoCodingScheme.setString(i++, registeredName);
            insertIntoCodingScheme.setString(i++, defaultLanguage_);
            insertIntoCodingScheme.setString(i++, representsVersion);
            DBUtility.setBooleanOnPreparedStatment(insertIntoCodingScheme, i++, new Boolean(isNative));
            insertIntoCodingScheme.setInt(i++, approxNumConcepts);
            if (sqlTableUtility_.getSQLTableConstants().supports2006Model())
            {
                DBUtility.setBooleanOnPreparedStatment(insertIntoCodingScheme, i++, null);
                DBUtility.setBooleanOnPreparedStatment(insertIntoCodingScheme, i++, null);
                DBUtility.setBooleanOnPreparedStatment(insertIntoCodingScheme, i++, null);
            }
            insertIntoCodingScheme.setString(i++, entityDescription);
            insertIntoCodingScheme.setString(i++, copyright);
            
            insertIntoCodingScheme.executeUpdate();

            messages_.info("Loaded 1 coding scheme");

            loadCodingSchemeMultiAttributes(UMLSCodingSchemeName, codingSchemeName);
            loadCodingSchemeSupportedAttributes(UMLSCodingSchemeName, codingSchemeName);

            loadConcepts(UMLSCodingSchemeName, codingSchemeName);
            loadRelations(UMLSCodingSchemeName, codingSchemeName);
            loadContexts(UMLSCodingSchemeName, codingSchemeName);

            buildRootNodes(codingSchemeName);
        }
        results.close();
        getCodingSchemeInfo.close();
    }

    /*
     * load the codingSchemeMultiAttributes table
     */
    private void loadCodingSchemeMultiAttributes(String UMLSCodingSchemeName, String codingSchemeName)
            throws SQLException
    {
        messages_.info("loading coding scheme multi attributes");
        // things like localName - source (which has a locatorInfo)

        insertIntoCodingSchemeMultiAttributes.setString(1, codingSchemeName);
        insertIntoCodingSchemeMultiAttributes.setString(2, "source");
        insertIntoCodingSchemeMultiAttributes.setString(3, "UMLS - " + getSourceVersionString("Unknown version of the UMLS"));
        if (stc_.supports2006Model())
        {
            insertIntoCodingSchemeMultiAttributes.setString(4, "");
            insertIntoCodingSchemeMultiAttributes.setString(5, "");
        }
        
        insertIntoCodingSchemeMultiAttributes.executeUpdate();

        insertIntoCodingSchemeMultiAttributes.setString(1, codingSchemeName);
        insertIntoCodingSchemeMultiAttributes.setString(2, "localName");
        insertIntoCodingSchemeMultiAttributes.setString(3, codingSchemeName);
        if (stc_.supports2006Model())
        {
            insertIntoCodingSchemeMultiAttributes.setString(4, "");
            insertIntoCodingSchemeMultiAttributes.setString(5, "");
        }

        insertIntoCodingSchemeMultiAttributes.executeUpdate();

        String iso = getISOString(UMLSCodingSchemeName);
        insertIntoCodingSchemeMultiAttributes.setString(1, codingSchemeName);
        insertIntoCodingSchemeMultiAttributes.setString(2, "localName");
        insertIntoCodingSchemeMultiAttributes.setString(3, iso);
        if (stc_.supports2006Model())
        {
            insertIntoCodingSchemeMultiAttributes.setString(4, "");
            insertIntoCodingSchemeMultiAttributes.setString(5, "");
        }

        insertIntoCodingSchemeMultiAttributes.executeUpdate();
        
        if (iso.startsWith("urn:iso:"))
        {

            insertIntoCodingSchemeMultiAttributes.setString(1, codingSchemeName);
            insertIntoCodingSchemeMultiAttributes.setString(2, "localName");
            insertIntoCodingSchemeMultiAttributes.setString(3, iso.substring("urn:iso:".length()));
            if (stc_.supports2006Model())
            {
                insertIntoCodingSchemeMultiAttributes.setString(4, "");
                insertIntoCodingSchemeMultiAttributes.setString(5, "");
            }

            insertIntoCodingSchemeMultiAttributes.executeUpdate();
        }
    }

    /*
     * load the codingSchemeSupported Attributes table
     */
    private void loadCodingSchemeSupportedAttributes(String UMLSCodingSchemeName, String codingSchemeName)
            throws SQLException
    {
        messages_.info("loading supported attributes");

        // language - done later

        // format
        // arbitrary choices
        insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "Format", "text/plain", null, null, null, null);
        insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "Format", "text/html", null, null, null, null);
        insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "Format", "text/xml", null, null, null, null);

        // property - these are loaded after the concepts.
        // property qualifiers
        supportedPropertyQualifiers_.add("HCD");
        loadSupportedPropertyQualifiers(codingSchemeName);

        // codingScheme
        // loaded later

        // source
        // currently none the way we load it.

        // association
        populateSupportedAssociations(UMLSCodingSchemeName);
        for (int i = 0; i < supportedAssociations_.length; i++)
        {
            insertIntoCodingSchemeSupportedAttributes(
                                                      codingSchemeName,
                                                      "Association",
                                                      (supportedAssociations_[i]).name,
                                                      null, null, null, null);
        }
        messages_.info("added " + supportedAssociations_.length + " associations");

        // associationQualifier
        supportedAssociationQualifiers_.add("HCD");
        loadSupportedAssociationQualifiers(codingSchemeName);

        // conceptStatus
        // TODO dig these out of the history - (but not now)

        // data type
        // no data types.

        // representationForm - from the atn codes in mrsat
        Enumeration values = atnRepresentationalMap_.elements();
        int i = 0;
        while (values.hasMoreElements())
        {
            String var = (String) values.nextElement();
            insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "RepresentationalForm", var, null, null, null, null);
            i++;
        }
        // more representationalForm - these are for tty codes in mrconso
        values = mrconsoRepresentationalMap_.elements();
        while (values.hasMoreElements())
        {
            String var = (String) values.nextElement();
            insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "RepresentationalForm", var, null, null, null, null);
            i++;
        }
        messages_.info("Loaded " + i + " representational forms");
    }

    private void populateSupportedAssociations(String UMLSCodingSchemeName) throws SQLException
	{
	    messages_.info("Getting the descriptive associations");
	    PreparedStatement getRelations = umlsConnection_
	            .prepareStatement("SELECT DISTINCT RELA, DIR FROM MRREL WHERE SAB = ?");
	
	    getRelations.setString(1, UMLSCodingSchemeName);
	
	    ResultSet relations = getRelations.executeQuery();
	    Hashtable relationsHolder = new Hashtable();
	
	    while (relations.next())
	    {
	        String temp = relations.getString("RELA");
	        String tempDirFlag = relations.getString("DIR");
	        if (temp == null || temp.length() == 0)
	        {
	            continue;
	        }
	        mapSupportedAssociationsHelper(temp, UMLSCodingSchemeName, tempDirFlag, "RELA", relationsHolder);
	    }
	
	    messages_.info("Getting the base umls associations");
	    getRelations = umlsConnection_
	            .prepareStatement("SELECT DISTINCT REL, DIR FROM MRREL WHERE SAB = ?");
	    getRelations.setString(1, UMLSCodingSchemeName);
	
	    relations = getRelations.executeQuery();
	    while (relations.next())
	    {
	        String temp = relations.getString("REL");
	        String tempDirFlag = relations.getString("DIR");
	        if (temp == null || temp.length() == 0)
	        {
	            continue;
	        }
	        mapSupportedAssociationsHelper(temp, UMLSCodingSchemeName, tempDirFlag, "REL", relationsHolder);
	    }
	    getRelations.close();
	
	    supportedAssociations_ = new Association[relationsHolder.size()];
	    Enumeration elements = relationsHolder.elements();
	    int i = 0;
	    while (elements.hasMoreElements())
	    {
	        Association temp = (Association) elements.nextElement();
	        supportedAssociations_[i++] = temp;
	    }
	}

	/*
     * load the concepts for a coding scheme
     */
    private void loadConcepts(String UMLSCodingSchemeName, String codingSchemeName) throws SQLException
    {
        messages_.info("loading concepts - getting a total count");

        PreparedStatement getCodingSchemeInfo = umlsConnection2_
                .prepareStatement("SELECT  COUNT(*) as cnt FROM MRCONSO WHERE SAB = ?");

        getCodingSchemeInfo.setString(1, UMLSCodingSchemeName);

        ResultSet results = getCodingSchemeInfo.executeQuery();

        results.next();
        int total = results.getInt("cnt");

        results.close();
        getCodingSchemeInfo.close();

        int start = 0;
        String lastCode = null;
        ArrayList conceptPresentations = new ArrayList();
        int codeCount = 0;
        int rowCount = 0;

        // collect them all into batches of concept codes. when the code changes, load that code
        while (start < total)
        {
            messages_.info("Fetching a batch of results");
            getCodingSchemeInfo = umlsConnection2_
                    .prepareStatement(umlsSqlModifier_
                            .modifySQL("SELECT  LAT, CODE, CUI, LUI, SUI, AUI, TS, STT, STR, TTY FROM MRCONSO WHERE SAB = ? ORDER BY {BINARY} CODE {LIMIT}"));

            getCodingSchemeInfo.setString(1, UMLSCodingSchemeName);
            
            //mysql doesn't stream results - the {LIMIT above and this is for getting limits on mysql code}
            if (umlsSqlModifier_.getDatabaseType().equals("MySQL"))
            {
                getCodingSchemeInfo.setInt(2, start);
                getCodingSchemeInfo.setInt(3, batchSize);
                start += batchSize;
            }
            else if (umlsSqlModifier_.getDatabaseType().equals("PostgreSQL"))
            {
                //postgres properly streams results, we can just set the fetch size, and only loop once
                getCodingSchemeInfo.setFetchSize(batchSize);
                umlsConnection2_.setAutoCommit(false);
                start = total;
            }
            else
            {
                start = total;
            }
            
            results = getCodingSchemeInfo.executeQuery();

            log.debug("query finished, processing results");

            while (results.next())
            {
                rowCount++;
                // store all the data from this row.

                ConceptPresentation temp = new ConceptPresentation();
                temp.conceptCode = results.getString("CODE");
                temp.cui = results.getString("CUI");
                
                //there are some concepts that just have "NOCODE" as the code... need to make them unique.
                if (temp.conceptCode.equals("NOCODE"))
                {
                    temp.conceptCode = temp.conceptCode + "-" + temp.cui;
                }

                if (lastCode != null && !temp.conceptCode.equals(lastCode))
                {
                    loadConcept(UMLSCodingSchemeName, codingSchemeName, (ConceptPresentation[]) conceptPresentations
                            .toArray(new ConceptPresentation[conceptPresentations.size()]));

                    conceptPresentations.clear();
                    commentCounter_ = 1;
                    presentationCounter_ = 1;
                    instructionCounter_ = 1;
                    definitionCounter_ = 1;
                    propertyCounter_ = 1;
                    cuiCounter_ = 1;
                    semTypeCounter_ = 1;

                    if (codeCount % 10 == 0)
                    {
                        messages_.busy();
                    }

                    codeCount++;
                    if (codeCount % 1000 == 0)
                    {
                        messages_.info("On row "
                                + rowCount
                                + " out of "
                                + total
                                + " rows - found "
                                + codeCount
                                + " concepts");
                    }
                }

                temp.language = results.getString("LAT");
                

                if (UMLSCodingSchemeName.equals("SNOMEDCT"))
                {
                    // snomed has a different way of recording the language of the presentation...
                    String snomedLanguage = getSnomedLanguageForPresentation(results.getString("CUI"), results
                            .getString("LUI"), results.getString("SUI"));
                    if (snomedLanguage != null && snomedLanguage.length() > 0)
                    {
                        temp.language = snomedLanguage;
                    }
                }
                // add for populating the supportedLanguages later
                supportedLanguages_.add(temp.language);

                temp.presentationFormat = "text/plain";
                temp.TTY = results.getString("TTY");
                temp.representationForm = temp.TTY;
                temp.value = results.getString("STR");
                temp.TS = results.getString("TS");
                temp.STT = results.getString("STT");
                temp.AUI = results.getString("AUI");
                temp.isPreferred = new Boolean(false);

                // See if there is a "better" value for the representationalForm mapping
                String repFormMap = (String) mrconsoRepresentationalMap_.get(temp.representationForm);
                if (repFormMap != null)
                {
                    temp.representationForm = repFormMap;
                }
                
                lastCode = temp.conceptCode;
                conceptPresentations.add(temp);
            }
            results.close();
        }

        // load the last one
        if (lastCode != null)
        {
            //need to add this, just once - assuming there were codes, so this is a good place for it.
            supportedPropertyTypes_.add("textualPresentation");
            loadConcept(UMLSCodingSchemeName, codingSchemeName, (ConceptPresentation[]) conceptPresentations
                    .toArray(new ConceptPresentation[conceptPresentations.size()]));

            codeCount++;
        }

        messages_.info("Loaded " + codeCount + " concepts from  " + rowCount + " rows");
        updateApproxNumberOfConcepts(codeCount, codingSchemeName);
        loadSupportedProperties(codingSchemeName);
        loadSupportedLanguages(codingSchemeName);
        getCodingSchemeInfo.close();
    }

    /*
     * Helper method to load all the properties and presentations of a concept
     */
    private void loadConcept(String UMLSCodingSchemeName, String codingSchemeName,
            ConceptPresentation[] conceptPresentations) throws SQLException
    {
        // order them - this will organize them by language, and put the "best" presentation
        // first for each language. default language will be the at the very top.
        Arrays.sort(conceptPresentations, new ConceptPresentationSorter());

        // now set each isPreferred to true for the first entry of each language
        String prevLang = conceptPresentations[0].language;
        conceptPresentations[0].isPreferred = new Boolean(true);
        for (int i = 1; i < conceptPresentations.length; i++)
        {
            if (!conceptPresentations[i].language.equals(prevLang))
            {
                conceptPresentations[i].isPreferred = new Boolean(true);
            }
            prevLang = conceptPresentations[i].language;
        }

        // load the entry into the concepts table - must be at least 1 row. the first one will
        // be the best text presentation to use here.
        try
        {
            addConceptToConcepts(codingSchemeName, conceptPresentations[0].conceptCode, null, null, null, new Boolean(true),
                                 "Active", new Boolean(false), conceptPresentations[0].value);
        }
        catch (SQLException e)
        {
            // Recover from a failure adding a new code. Likely caused by trying
            // to insert a duplicate code - some code systems are case sensitive - and
            // most of our databases aren't.
            log.error("Problem inserting new code " + conceptPresentations[0].conceptCode, e);
            messages_.info("ERROR - Problem inserting new code " + conceptPresentations[0].conceptCode);
            return;
        }
        
        LRUMap TValues = new LRUMap(20);

        // load its properties
        for (int i = 0; i < conceptPresentations.length; i++)
        {
        	String propertyId = "T-" + presentationCounter_++;
            addToConceptProperty(codingSchemeName, conceptPresentations[i].conceptCode, propertyId,
                                 "presentation", "textualPresentation", conceptPresentations[i].language,
                                 conceptPresentations[i].presentationFormat, conceptPresentations[i].isPreferred,
                                 null, null, conceptPresentations[i].representationForm, conceptPresentations[i].value);
            TValues.put(conceptPresentations[i].AUI, propertyId);
        }

        // load the definitions
        addToDefinitions(UMLSCodingSchemeName, conceptPresentations[0].cui, codingSchemeName,
                         conceptPresentations[0].conceptCode);

        // load the other properties
        loadOtherProperties(UMLSCodingSchemeName, conceptPresentations[0].cui, codingSchemeName,
                            conceptPresentations[0].conceptCode, TValues);
        
        // TODO dig the history out - add the modVersion (not now)
        // addConceptToConceptsMultiAttributes(codingSchemeName, conceptCode, "modVersion", "");
        
        // this is where the usageContext would be loaded - but we don't have any.
        // also, if we were to load the UMLS as a coding scheme (instead of individual coding schemes) this is
        // where we would put the source information
        // addConceptToConceptPropertyMultiAttributes(codingSchemeName, conceptCode, propertyId, "", "", "",
        // "");
    }

    /*
	 * This is for mapping the MRSAT table into comments, instructions, presentations, and just plain old properties.
	 */
	private void loadOtherProperties(String UMLSCodingSchemeName, String cui, String codingSchemeName,
	        String conceptCode, LRUMap TValues) throws SQLException
	{
	    // Drop the CUI in.
	    supportedPropertyTypes_.add("UMLS_CUI");
	    addToConceptProperty(codingSchemeName, conceptCode, "CUI-" + cuiCounter_++, "property", "UMLS_CUI", null, null, null,
	                         null, null, null, cui);
	
	    //get the semantic type(s)
	    supportedPropertyTypes_.add("Semantic_Type");
	    PreparedStatement getSemTypes = umlsConnection_.prepareStatement("SELECT STY FROM MRSTY WHERE CUI = ?");
	    getSemTypes.setString(1, cui);
	    
	    ResultSet semTypes = getSemTypes.executeQuery();
	    while (semTypes.next())
	    {
	        addToConceptProperty(codingSchemeName, conceptCode, "SemType-" + semTypeCounter_++, "property", "Semantic_Type", null, "text/plain", null,
	                             null, null, null, semTypes.getString("STY"));
	    }
	    
	    getSemTypes.close();
	    
	    //load the other properties
	    PreparedStatement getProperties = umlsConnection_
	            .prepareStatement("SELECT ATN, ATV, STYPE, METAUI FROM MRSAT WHERE CODE = ? AND SAB = ?");
	
	    getProperties.setString(1, conceptCode);
	    getProperties.setString(2, UMLSCodingSchemeName);
	
	    ResultSet properties = getProperties.executeQuery();
	    
	    ArrayList auiProps = new ArrayList();
	
	    while (properties.next())
	    {
	        String propertyFlag = properties.getString("ATN");
	        String propertyValue = properties.getString("ATV");
	        String propertyStype = properties.getString("STYPE");
	        String propertyMetaui = properties.getString("METAUI");
	
	        String propertyId = null;
	        String propertyName = null;
	        String propertyType = null;
	        String representationalForm = null;
	        Boolean isPreferred = null;
	
	        if (mrsatMap_.get(propertyFlag) != null &&
	        		(propertyStype.equalsIgnoreCase("CODE") ||
	        		 propertyStype.equalsIgnoreCase("SCUI")||
	        		 propertyStype.equalsIgnoreCase("SDUI")))
	        {
	            Integer type = (Integer) mrsatMap_.get(propertyFlag);
	            if (type.intValue() == SKIP.intValue())
	            {
	                continue;
	            }
	            else if (type.intValue() == COMMENT.intValue())
	            {
	                propertyId = "C-" + commentCounter_++;
	                propertyType = "comment";
	                propertyName = "comment";
	                isPreferred = new Boolean(false);
	            }
	            else if (type.intValue() == INSTRUCTION.intValue())
	            {
	                propertyId = "I-" + instructionCounter_++;
	                propertyType = "instruction";
	                propertyName = "instruction";
	                isPreferred = new Boolean(false);
	            }
	            else if (type.intValue() == PRESENTATION.intValue())
	            {
	                propertyId = "T-" + presentationCounter_++;
	                propertyType = "presentation";
	                propertyName = "textualPresentation";
	                TValues.put(propertyMetaui, propertyId);
	                isPreferred = new Boolean(false);
	            }
	        }
	        else if (propertyStype.equalsIgnoreCase("SAUI")) {
	        	auiProps.add(new AUI_PROP(propertyFlag, propertyValue, propertyStype, propertyMetaui));
	        	continue;
	        }
	        else
	        {
	            propertyId = "P-" + propertyCounter_++;
	            propertyType = "property";
	            propertyName = propertyFlag;
	        }
	        representationalForm = (String) atnRepresentationalMap_.get(propertyFlag);
	        supportedPropertyTypes_.add(propertyType);
	        addToConceptProperty(codingSchemeName, conceptCode, propertyId, propertyType, propertyName, null, "text/plain", 
	                             isPreferred, null, null, representationalForm, propertyValue);
	    }
	
	    Iterator aPIter = auiProps.iterator();
	    while (aPIter.hasNext()) {
	    	AUI_PROP ap = (AUI_PROP) aPIter.next();
	    	String apPropertyId = (String) TValues.get(ap.metaui);
	    	if (apPropertyId == null)
	    		continue;
	    	if (!supportedPropertyQualifiers_.contains(ap.atn))
	    		supportedPropertyQualifiers_.add(ap.atn);
	    	addConceptToConceptPropertyMultiAttributes(codingSchemeName, conceptCode, apPropertyId, "qualifier", ap.atn, ap.atv, "");
	    	
	    }
	    
	    getProperties.close();
	}

	/*
	 * Load the definitions for a concept
	 */
	private void addToDefinitions(String UMLSCodingSchemeName, String cui, String codingSchemeName, String conceptCode) throws SQLException
	{
	    PreparedStatement getDefinitions = umlsConnection_
	            .prepareStatement("SELECT DEF FROM MRDEF WHERE CUI = ? AND SAB = ?");
	
	    getDefinitions.setString(1, cui);
	    getDefinitions.setString(2, UMLSCodingSchemeName);
	
	    ResultSet definitions = getDefinitions.executeQuery();
	    boolean isPreferred = true;
	    while (definitions.next())
	    {
	        supportedPropertyTypes_.add("definition");
	        addToConceptProperty(codingSchemeName, conceptCode, "D-" + definitionCounter_++, "definition", "definition", null,
	                             null, new Boolean(isPreferred), null, null, null, definitions
	                                     .getString("DEF"));
	        isPreferred = false;
	        // umls doesn't seem to have the notion of a preferred definition, so I'll just set the first one to
	        // preferred.
	    }
	    getDefinitions.close();
	}

	private String getSnomedLanguageForPresentation(String cui, String lui, String sui) throws SQLException
	{
	    getSnomedLanguage_.setString(1, cui);
	    getSnomedLanguage_.setString(2, lui);
	    getSnomedLanguage_.setString(3, sui);
	
	    ResultSet results = getSnomedLanguage_.executeQuery();
	    String result = null;
	    if (results.next())
	    {
	        result = results.getString("ATV");
	    }
	    results.close();
	    return result;
	}

	private void loadRelations(String UMLSCodingSchemeName, String codingSchemeName) throws SQLException
	{
		// Sort the associations, loading RELA's first and keeping
		// non-directional names (potentially qualified by different SABs)
		// together.
		Arrays.sort(supportedAssociations_, new AssociationSorter());

		for (int i = 0; i < supportedAssociations_.length; i++)
		{
			Association assoc = (Association) supportedAssociations_[i];
			if (lastAssociation == null || !lastAssociation.equals(assoc.name))
			{
				// Moving on to a new association name.
				lastAssociation = assoc.name;
				// Clear the hashtable used to track source/target combinations loaded.
				alreadyLoadedAssociations.clear();          
			}

			int count = loadRelationsHelper(assoc, UMLSCodingSchemeName, codingSchemeName);
			if (count == 0)
			{
				messages_.info("No relations were found for " + assoc.toShortString());
				log.warn("No relations were found for " + assoc.toShortString());
			}
		}

		// populate the supported coding schemes, now that we know what they are.
		Iterator values = supportedCodingSchemes_.iterator();
		int i = 0;
		while (values.hasNext())
		{
			String var = (String) values.next();
			String urn = getISOString(mapCodeSystemNameToSAB(var));
			insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "CodingScheme", var, urn, null, null, null);
			i++;
		}

		messages_.info("Loaded " + i + " supported coding schemes");
	}

	private int loadRelationsHelper(Association assoc, String UMLSCodingSchemeName, String codingSchemeName)
            throws SQLException
    {
		PreparedStatement getRRFRelationsCount = umlsConnection2_.prepareStatement(
			"SELECT COUNT(*) AS cnt FROM MRREL WHERE "
				+ assoc.rrfField + " = ? AND SAB = ?");
		// Note: ordered to enable a cache to not load duplicates.
		PreparedStatement getRRFRelations = umlsConnection2_.prepareStatement(umlsSqlModifier_.modifySQL(
			"SELECT  CUI1, AUI1, CUI2, AUI2 FROM MRREL WHERE "
				+ assoc.rrfField + " = ? AND SAB = ? ORDER BY CUI1 {LIMIT}"));
		int count = 0;
		
		try
		{
			int start = 0;
			int total = 0;
			int dupeCount = 0;

			getRRFRelationsCount.setString(1, assoc.rrfName);
			getRRFRelationsCount.setString(2, UMLSCodingSchemeName);

			messages_.info("Counting relations for " + assoc.name);
			ResultSet results = getRRFRelationsCount.executeQuery();
			results.next();
			total = results.getInt("cnt");
			results.close();

			while (start < total)
			{
				getRRFRelations.setString(1, assoc.rrfName);
				getRRFRelations.setString(2, UMLSCodingSchemeName);

				// mysql doesn't stream results - the {LIMIT above and this is for getting limits on mysql code}
				if (umlsSqlModifier_.getDatabaseType().equals("MySQL"))
				{
					getRRFRelations.setInt(3, start);
					getRRFRelations.setInt(4, batchSize * 3);
					start += batchSize * 3;
				}
				// postgres properly streams results, we can just set the fetch size, and only loop once
				else if (umlsSqlModifier_.getDatabaseType().equals("PostgreSQL"))
				{
					getRRFRelations.setFetchSize(batchSize * 3);
					umlsConnection2_.setAutoCommit(false);
					start = total;
				}
				else
				{
					start = total;
				}

				messages_.info("Getting a batch of relations for " + assoc.rrfName);
				results = getRRFRelations.executeQuery();

				while (results.next())
				{
					// If the source association definition is reversed, we need to flip
					// the source and target concepts accordingly.
					boolean isReversed = assoc.rrfSourceDirectionalityReversed;
					String cui = results.getString(isReversed ? "CUI2" : "CUI1");
					String aui = results.getString(isReversed ? "AUI2" : "AUI1");

					CodeHolder[] sourceCode = mapCUIToCode(cui, aui, UMLSCodingSchemeName);

					if (sourceCode == null || sourceCode.length == 0)
					{
						log.error("Could not find sourceCode for " + cui + " : " + aui);
						continue;
					}

					cui = results.getString((isReversed ? "CUI1" : "CUI2"));
					aui = results.getString((isReversed ? "AUI1" : "AUI2"));

					CodeHolder[] targetCode = mapCUIToCode(cui, aui, UMLSCodingSchemeName);

					if (targetCode == null || targetCode.length == 0)
					{
						log.error("Could not find targetCode for " + cui + " : " + aui);
						continue;
					}

					/*
					 * Most terminologies have a one to one mapping from cui - aui pairs to unique codes.  
					 * So there will only be one source and one target concept code.  RXNorm and MTH, however, 
					 * don't always provide AUI's.  When no AUI's are provided, it is possible to have the CUI's
					 * map to more than one source or target code.
					 */
					for (int i = 0; i < sourceCode.length; i++)
					{
						for (int j = 0; j < targetCode.length; j++)
						{
							int tempval = addConceptAssociationToConceptsHelper(
									codingSchemeName, sourceCode[i],
									assoc, targetCode[j]);

							if (tempval > 0)
							{
								count++;
							}
							else if (tempval < 0)
							{
								dupeCount++;
							}
						}
					}

					if (count % 100 == 0)
					{
						messages_.busy();
					}

					if (count % 10000 == 0)
					{
						messages_.info("Loaded " + count + " out of a possible total of " + total);
					}
				}
				results.close();
			}
			messages_.info("Loaded " + count + " out of a possible total of " + total);
			if (dupeCount > 0)
			{
				messages_.info("Encountered " + dupeCount + " duplicates");
			}
		}
		finally
		{
			getRRFRelations.close();
			getRRFRelationsCount.close();
		}
        return count;
    }
    
	/**
     * Add a code to code association.
	 * @param currentCodingScheme
	 * @param sourceCode
	 * @param association
	 * @param targetCode
     * @return 1 if the association was added; 0 if an error
     * occurred; -1 if the association was already defined.
	 * @throws SQLException
	 */
    private int addConceptAssociationToConceptsHelper(String currentCodingScheme, CodeHolder sourceCode,
            Association association, CodeHolder targetCode) throws SQLException
    {
        RelationType relationType = getRelationType(association.rrfField, sourceCode.codingScheme,
                                                    targetCode.codingScheme);

        if (!loadedRelations_.contains(currentCodingScheme + relationType.relation))
        {
            messages_.info("Adding the relations container to put the associations under.");

            addRelationToRelations(currentCodingScheme, relationType.relation, new Boolean(relationType.isNative),
                                   relationType.relationDescription);
            loadedRelations_.add(currentCodingScheme + relationType.relation);
        }

        if (!isLoaded(association, currentCodingScheme, relationType.relation))
        {
        	// If hierarchical, indicate transitivity.
            Boolean transitive = new Boolean(isHierarchicalAssociation(association, currentCodingScheme));
            messages_.info("Adding association: " + association.name);
            addAssociationToAssociations(
            		currentCodingScheme, relationType.relation,
            		association.name, association.rrfName, association.rrfInverse, "",
            		new Boolean(true), transitive,
            		null, null, null, null, null, null, null, null, null,
            		"UMLS-defined relationships");
            markLoaded(association, currentCodingScheme, relationType.relation);
        }
        
        int returnVal = 0;

        // No need for association name here, since we always go through in a known order.
        String key = sourceCode.codingScheme + ":" + sourceCode.code + ":" + targetCode.codingScheme + ":" + targetCode.code;

        // Already registered this one?
        if (!alreadyLoadedAssociations.contains(key))
        {
        	// Try optimistic insert.  On error, check for duplicate just
        	// in case there was some failure to cache the key.
        	// Otherwise, throw the original err.
        	try {
		        addConceptAssociationToConcepts(
                		currentCodingScheme,
                		sourceCode.codingScheme, sourceCode.code,
                        relationType.relation, association.name,
                        targetCode.codingScheme, targetCode.code,
                        null, null, null);
                alreadyLoadedAssociations.add(key);
                returnVal = 1;
        	} catch (SQLException sqle) {
				ResultSet rs = null;
				getAssocInstance_.setString(1, currentCodingScheme);
				getAssocInstance_.setString(2, relationType.relation);
				getAssocInstance_.setString(3, association.name);
				getAssocInstance_.setString(4, sourceCode.codingScheme);
				getAssocInstance_.setString(5, sourceCode.code);
				getAssocInstance_.setString(6, targetCode.codingScheme);
				getAssocInstance_.setString(7, targetCode.code);
				try {
					rs = getAssocInstance_.executeQuery();
					if (rs.next()) {
		                alreadyLoadedAssociations.add(key);
		                returnVal = -1;
					} else {
						throw sqle;
					}
				} catch (Exception e) {
        			log.warn("Association instance not inserted.", sqle);
				} finally {
					if (rs != null)
						rs.close();
				}
            }
        }
        else {
        	returnVal = -1;
        }
        return returnVal;
    }
    
    private RelationType getRelationType(String relationSource, String sourceCodingScheme, String targetCodingScheme)
    {
        String relation;
        String relationDescription;
        boolean isNative;
        
        supportedCodingSchemes_.add(sourceCodingScheme);
        supportedCodingSchemes_.add(targetCodingScheme);
        
        // REL and RELA relations are split across relation containers to
        // separate standard UMLS relation names from source-specific labels.
        // However, with interpretation of NLM as curator both are tagged as
        // native (defined by curator).  This also serves to make default queries
        // against coded node sets and graphs more inclusive.
        if (relationSource.equals("REL"))
        {
            isNative = true;
            relation = STD_RELATIONS;
            relationDescription = "UMLS-defined relationships (e.g. RB, RN, CHD, PAR)";
         }
        else // 'RELA'
        {
            isNative = true;
            relation = SRC_RELATIONS;
            relationDescription = "Source-defined relationships.";
        }
 
        return new RelationType(relation, relationDescription, isNative);
    }
    
    /**
	 * Iterate MRHIER for context information to be added as Association Qualifications
	 * @param UMLSCodingSchemeName
	 * @param codingSchemeName
	 */
	private void loadContexts(String UMLSCodingSchemeName, String codingSchemeName) throws SQLException {
		/*
		 * Evaluate hierarchy processing option.
		 * If 0, no context entries in MRHIER are processed.
		 * If 1, only entries tagged with a context ID are evaluated.
		 * when absent.
		 */
		if (mrhierOpt_ == 0) {
			messages_.info("Bypassing MRHIER entries.");
			return;
		} else {
			messages_.info("Processing HCD-tagged MRHIER entries.");
		}
		boolean constructHCD = false;
		 
		 /*
		  * Create a temporary view (that will be destroyed at the end 
		  * of the method) to speed up the queries in this method
		  */
		 try {
	     	 messages_.info("loading contexts - getting a total count");
	    	 
	    	 String getCodingSchemeInfoSQL = "SELECT COUNT(1) as cnt FROM MRHIER" +
	    	 		" WHERE SAB = ? ";
	    	 if (!constructHCD) {
	    		 getCodingSchemeInfoSQL = getCodingSchemeInfoSQL +
	 				" AND HCD != ? ";
	    	 }
		 
	    	 PreparedStatement getCodingSchemeInfo = umlsConnection2_
	             	.prepareStatement(umlsSqlModifier_.modifySQL(getCodingSchemeInfoSQL));
	    	 ResultSet results = null;
	
	    	 getCodingSchemeInfo.setString(1, UMLSCodingSchemeName);
	    	 // this is weird, but its the easiest way to get the 
	    	 //  appropriate quotes around NULL
	    	 if (!constructHCD)
	    		 getCodingSchemeInfo.setString(2, "NULL");
	    	 int total = 0;
	    	 try
	    	 {
	        	 results = getCodingSchemeInfo.executeQuery();
	        	 
	        	 results.next();
	        	 total = results.getInt("cnt");
	
	        	 results.close();
	    	 }
	    	 finally
	    	 {
	    		 getCodingSchemeInfo.close();
	    	 }
	
	    	 int start = 0;
	    	 int contextCount = 0;
	    	 int rowCount = 0;
	     
	    	 getCodingSchemeInfoSQL = 
	    		 "SELECT MRHIER.CUI, MRHIER.AUI, MRHIER.PTR, MRHIER.HCD, " +
	    		 " MRHIER.SAB, MRHIER.RELA, MRHIER.CXN " +
	    		 " FROM MRHIER " +
	    		 " WHERE SAB = ? ";
	    	 if (!constructHCD) {
	    		 getCodingSchemeInfoSQL = getCodingSchemeInfoSQL +
	 				" AND HCD != ? ";
	    	 }
	    	 getCodingSchemeInfoSQL = getCodingSchemeInfoSQL + " ORDER BY {BINARY} MRHIER.CUI {LIMIT}";
	    	 getCodingSchemeInfo = umlsConnection2_
	             	.prepareStatement(umlsSqlModifier_
	             			.modifySQL(getCodingSchemeInfoSQL));
	
	    	 // Count var usage to make sure things are placed correctly
	    	 int boundVar = 1;
	    	 getCodingSchemeInfo.setString(boundVar++, UMLSCodingSchemeName);
	    	 // this is weird, but its the easiest way to get the 
	    	 //  appropriate quotes around NULL
	    	 if (!constructHCD)
	    		 getCodingSchemeInfo.setString(boundVar++, "NULL");
	    	 try
	    	 {
	        	 // collect them all into batches of concept codes. when the code changes, load that code
	        	 while (start < total)
	        	 {
	        		 messages_.info("Fetching a batch of results");
	                          
	        		 if (umlsSqlModifier_.getDatabaseType().equals("MySQL"))
	        		 {
		        		 //mysql doesn't stream results - the {LIMIT above and this is for getting limits on mysql code}
	        			 getCodingSchemeInfo.setInt(boundVar, start);
	        			 getCodingSchemeInfo.setInt(boundVar+1, batchSize);
	        			 start += batchSize;
	        		 }
	        		 else if (umlsSqlModifier_.getDatabaseType().equals("PostgreSQL"))
	        		 {
	        			 //postgres properly streams results, we can just set the fetch size, and only loop once
	        			 getCodingSchemeInfo.setFetchSize(batchSize);
	        			 umlsConnection2_.setAutoCommit(false);
	        			 start = total;
	        		 }
	        		 else
	        		 {
	        			 start = total;
	        		 }
	             
	        		 results = getCodingSchemeInfo.executeQuery();
	        		 try
	        		 {
		        		 messages_.debug("query finished, processing results");
		
		        		 while (results.next())
		        		 {
		        			 rowCount++;
		        			 // store all the data from this row.
		                 
		        			 AssociationQualification aq = new AssociationQualification();
		        			 aq.codingSchemeName = codingSchemeName;
		        			 String cui = results.getString("CUI");
		        			 aq.sourceConceptAUI = results.getString("AUI");
		        			 aq.sourceConceptCode = mapCUIToCode(cui, aq.sourceConceptAUI, UMLSCodingSchemeName)[0].code;
		        			 aq.qualifierName = "HCD";
		        			 aq.qualifierValue = results.getString("HCD");
		        			 aq.pathToRoot = results.getString("PTR");                 
		                 
		        			 contextCount += loadContext(aq, constructHCD, results.getString("RELA"), contextCount);
		                 
		        			 if (rowCount % 10 == 0)
		        			 {
		        				 messages_.busy();
		        			 }
		
		        			 if (rowCount % 1000 == 0)
		        			 {
		        				 messages_.info("On context "
		        						 + rowCount
		        						 + " out of "
		        						 + total
		        						 + " - found "
		        						 + contextCount
		        						 + " contextual links");
		        			 }
		        		 }
	        		 }
		        	 finally
		        	 {
		        		results.close();
		        	 }
	        	 }
	        	 messages_.info("Loaded " + contextCount + " contextual links from  " + rowCount + " contexts");
	    	 }
	    	 finally
	    	 {
	    		 getCodingSchemeInfo.close();
	    	 }
	     }
	     finally {
	     }
	}

	/**
	 * Adds qualification to concepts and associations in the LexGrid repository.
	 * @param aq Qualification information from the UMLS source.
	 * @param constructHCD Indicates whether artificial context values should be
	 * constructed if not provided in the UMLS information.
	 * @param rela The relationship attribute defined by UMLS (can be empty or null).
	 * @param totalCount The total number of context links qualified previously.
	 * @return The number of contextual links qualified in the repository
	 * for the given UMLS info.
	 * @throws SQLException
	 */
	private int loadContext(AssociationQualification aq, boolean constructHCD, String rela, int totalCount) throws SQLException {
		// If a context identifier was assigned, use it.
		// If a context identifier is not assigned and the option to construct is enabled,
		// derive one based on the root concept code and path to root AUI values.
		int contextLinks = 0;
		String hcd = aq.qualifierValue;
		if (constructHCD && StringUtils.isBlank(hcd)
				&& StringUtils.isNotBlank(aq.pathToRoot) && StringUtils.isNotBlank(aq.sourceConceptCode))
		{
			MessageDigest md = getSHA1();
			md.reset();
			md.update(aq.pathToRoot.getBytes());
			hcd = String.valueOf(md.digest(aq.sourceConceptCode.getBytes()));
		}
		if (StringUtils.isBlank(hcd))
			return 0;
		
		// Iterate through the path to root and determine the codes for
		// participating AUIs.  We maintain a LRU cache of AUIs to codes to assist.
		// If the associated code is not in the cache, find and cache it here.
		ListOrderedMap orderedPtrAUIToCode = new ListOrderedMap();
	
		// Break up the path to root into AUIs ...
		String[] auis = aq.pathToRoot.split("\\.");
		if (auis.length > 0) {
	    	// Check the cache for each.  If not found, perform and cache the
			// AUI to code mapping.
		   	PreparedStatement getPTRCode =
		   		umlsConnection2_.prepareStatement("SELECT CODE FROM MRCONSO WHERE AUI = ?");
		   	try {
	    		String nextCode, nextAUI;
	        	for (int i = 0; i < auis.length; i++) {
	        		// Check for registered code in the cache.
	        		nextAUI = auis[i];
	           		nextCode = (String) auiToCodeCache_.get(nextAUI);
	           		
	           		// If not cached, perform lookup ...
	           		if (nextCode == null) {
		    			getPTRCode.setString(1, nextAUI);
	    	    		ResultSet ptrCodes = getPTRCode.executeQuery();
	    	    		int count = 0;
	    	    		try {
	    	    			while (ptrCodes.next()) {
	    	    				count++;
	    		    			nextCode = ptrCodes.getString(1);
	    		    		}
	    	    		} finally {
	    	    			ptrCodes.close();
	    	    		}
		    			// If one to one mapping (probably should always be, but doesn't
		    			// hurt to check), add to the cache for quick lookup later...
		    			if (count == 1)
			    			auiToCodeCache_.put(nextAUI, nextCode);
	           		}
	           		
	           		// Was it resolved?
	           		if (nextCode != null)
	           			orderedPtrAUIToCode.put(nextAUI, nextCode);
	        	}
		   	} finally {
		   		getPTRCode.close();
		   	}
		}
		// Ensure we have included the original AUI to code mapping from the
		// provided UMLS qualification info; inserted last as the root
		// of the path.
		orderedPtrAUIToCode.put(aq.sourceConceptAUI, aq.sourceConceptCode);
	
		///////////////////////////////////////////////////////////////////////
		// We have all the participating codes and AUIs. 
		// Add context qualifiers to the text presentation of each concept
		// based on code/AUI pairs.
		///////////////////////////////////////////////////////////////////////
		for (OrderedMapIterator omi = orderedPtrAUIToCode.orderedMapIterator(); omi.hasNext(); ) {
			omi.next();
			String aui = (String) omi.getKey();
			String code = (String) omi.getValue();
			if (code != null)
				qualifyConceptPresentation(code, aui, aq.codingSchemeName,
					aq.qualifierName, hcd);
		}
		
		///////////////////////////////////////////////////////////////////////
		// At this point we have taken care of all the concept qualifiers.
		// Now find and similarly tag each participating association link
		// between AUIs in the path to root chain.
		///////////////////////////////////////////////////////////////////////
		
		// Statements to find LexGrid association to concept mappings.
		// Check source to target (parent association as processed)
		// or target to source (child association as processed).
		
		// Honor the association specified in the MRHIER entry, if provided.
		// For example, the UMLS 'inverse_isa' is mapped on load to 'hasSubtype'
		// association name; account for that here.
		String assoc = mapRela(rela);
		
		// If a specific relation attribute (rela) was not provided, consider all relevant
		// hierarchical associations (including UMLS standard or source-specific names).
		String assocParam = StringUtils.isNotBlank(assoc)
			? '\''+assoc+'\'' : toCommaDelimitedWithQuotes(getHierAssocNames(aq.codingSchemeName));
	
		// Create statements to navigate both directions (up & down the contextual chain).
		PreparedStatement getRelationship_1 = 
			sqlConnection_.prepareStatement(
					new StringBuffer("SELECT multiAttributesKey, targetConceptCode, association, sourceConceptCode FROM ")
						.append(stc_.getTableName(SQLTableConstants.CONCEPT_ASSOCIATION_TO_CONCEPT))
						.append(" WHERE sourceConceptCode = ? AND targetConceptCode = ? AND")
						.append(" codingSchemeName = ? AND association IN (").append(assocParam).append(")")
						.toString());
		PreparedStatement getRelationship_2 = 
			sqlConnection_.prepareStatement(
					new StringBuffer("SELECT multiAttributesKey, targetConceptCode, association, sourceConceptCode FROM ")
						.append(stc_.getTableName(SQLTableConstants.CONCEPT_ASSOCIATION_TO_CONCEPT))
						.append(" WHERE targetConceptCode = ? AND sourceConceptCode = ? AND")
						.append(" codingSchemeName = ? AND association IN (").append(assocParam).append(")")
						.toString());
		
		// Statement to update a multi-attributes key for an association mapping.
		PreparedStatement updateMAK =
			sqlConnection_.prepareStatement(
					new StringBuffer("UPDATE ")
						.append(stc_.getTableName(SQLTableConstants.CONCEPT_ASSOCIATION_TO_CONCEPT))
						.append(" SET multiAttributesKey = ? WHERE codingSchemeName = ?")
						.append(" AND sourceConceptCode = ? AND targetConceptCode = ?")
						.append(" AND association = ?")
						.toString());
		
		// Locate and qualify each affected association link with the context ID ...
		try {
			PreparedStatement[] stmts = new PreparedStatement[] {getRelationship_1, getRelationship_2};
			for (int s = 0; s < stmts.length; s++) {
				PreparedStatement stmt = stmts[s];
		    	for (int i = orderedPtrAUIToCode.size()-1; i > 0; i--) {
		    		String code = (String)orderedPtrAUIToCode.getValue(i);
		    		String codePrev = (String)orderedPtrAUIToCode.getValue(i-1);
		    		stmt.setString(1, code);
		    		stmt.setString(2, codePrev);
		    		stmt.setString(3, aq.codingSchemeName);
		    	
		    		ResultSet results = stmt.executeQuery();
		    		try {
		    			// Iterate through all relevant association links ...
		    			while (results.next()) {
		    				String multiAttributesKey = results.getString("multiAttributesKey");
		    				String targetConceptCode = results.getString("targetConceptCode");
		    				String sourceConceptCode = results.getString("sourceConceptCode");
		    				String association = results.getString("association");
		    				
			    			// If there is no key to correlate to the multi-attributes table,
		    				// construct and add now.
		    				if (multiAttributesKey == null) {
		    					StringBuffer key = new StringBuffer()
		    						.append(System.currentTimeMillis())
		    						.append((int)Math.floor((Math.random() * 100000)))
		    						.append(totalCount);
		    					multiAttributesKey = key.substring(0, Math.min(50, key.length()));
		    					updateMAK.setString(1, multiAttributesKey);
		    					updateMAK.setString(2, aq.codingSchemeName);
		    					updateMAK.setString(3, sourceConceptCode);
		    					updateMAK.setString(4, targetConceptCode);
		    					updateMAK.setString(5, association);
		    					updateMAK.execute();
		    				}
		    				
		    				// Add a context qualifier to the multi-attributes table.
		    				try { 
		    					addConceptAssociationQualifierToConceptAssociation(aq.codingSchemeName, multiAttributesKey, aq.qualifierName, hcd);
			    				contextLinks++;
		    				}
		    				catch (SQLException e) {
		    					// Because we qualify all relationships along the PTR and 
		    					// the HCD is identical for siblings at the same PTR some
		    					// exceptions with regards to identical keys will come up.
		    					
		    					// We try to bypass altogether if the message indicates duplication.
		    					// However, message text can vary based on the database engine.
		    					// Rather than exit in error, log the message and continue.
		    					if (!e.getMessage().contains("Duplicate")) {
		    						messages_.warn("Unable to add context qualifier to association.", e);
		    					}
		    				}
		    			}
		    		} finally { 
		    			results.close();
		    		}
		    	}
			}
		} finally { 
			updateMAK.close();
	   		getRelationship_1.close();
	   		getRelationship_2.close();
	   	}
		return contextLinks;
	}

	private CodeHolder[] mapCUIToCode(String cui, String aui, String sab) throws SQLException
	{
	    if (aui == null || aui.length() == 0)
	    {
	        return mapCUIToCodeNoAUI(cui, sab);
	    }
	    else
	    {
	        return mapCUIToCodeWithAUI(cui, aui);
	    }
	}

	private CodeHolder[] mapCUIToCodeNoAUI(String cui, String sab) throws SQLException
	{
	    CodeHolder[] resultToReturn = (CodeHolder[]) cuiToCodeCache_.get(cui + ":null:" + sab);
	    if (resultToReturn == null)
	    {
	        PreparedStatement statement = null;
	        
	        statement = getCodeForCUINoAUI_;
	        statement.setString(1, cui);
	        
	        if (sab.toLowerCase().equals("rxnorm"))
	        {
	            //if we are loading rxnorm, restrict matches to mrconso rows that are from rxnorm.
	            statement.setString(2, sab);
	        }
	        else
	        {
	            //not rxnorm, match from every sab.
	            statement.setString(2, "%");
	        }
	         
	        ArrayList temp = new ArrayList();
	        ResultSet results = statement.executeQuery();
	        
	        while (results.next())
	        {
	            CodeHolder tempResult = new CodeHolder();
	            tempResult.code = results.getString("CODE");
	            // there are some concepts that just have "NOCODE" as the code... need to make them unique.
	            if (tempResult.code.equals("NOCODE"))
	            {
	                tempResult.code = tempResult.code + "-" + cui;
	            }
	            tempResult.codingScheme = mapSABToCodeSystemName(results.getString("SAB"));
	            temp.add(tempResult);
	        }
	        
	        results.close();
	        
	        if (temp.size() == 0)
	        {
	            log.warn("Unable to map CUI to CODE - " + cui + " (no aui) sab - " + sab);
	        }
	        
	        resultToReturn = (CodeHolder[])temp.toArray(new CodeHolder[temp.size()]);
	        
	        cuiToCodeCache_.put(cui + ":null:" + sab, resultToReturn);
	    }
	    
	    return resultToReturn;
	}

	private CodeHolder[] mapCUIToCodeWithAUI(String cui, String aui) throws SQLException
	{
	    CodeHolder result = (CodeHolder) cuiToCodeCache_.get(cui + ":" + aui);
	    if (result == null)
	    {
	        PreparedStatement statement = null;
	
	        statement = getCodeForCUI_;
	        statement.setString(1, cui);
	        statement.setString(2, aui);
	
	        ResultSet results = statement.executeQuery();
	        if (results.next())
	        {
	            result = new CodeHolder();
	            result.code = results.getString("CODE");
	            // there are some concepts that just have "NOCODE" as the code... need to make them unique.
	            if (result.code.equals("NOCODE"))
	            {
	                result.code = result.code + "-" + cui;
	            }
	            result.codingScheme = mapSABToCodeSystemName(results.getString("SAB"));
	
	            cuiToCodeCache_.put(cui + ":" + aui, result);
	        }
	        else
	        {
	            log.warn("Unable to map CUI to CODE - " + cui + " (aui) - " + aui);
	        }
	        if (results.next())
	        {
	            log.error("INVALID ASSUMPTION - CUI-AUI is not unique");
	            log.error("cui " + cui + " aui " + aui);
	        }
	        results.close();
	    }
	    return result != null ? new CodeHolder[] {result} : null;
	}

	/**
	 * Add context qualifiers to text presentations for matching concepts.
	 * @param code The code of the concept to qualify.
	 * @param aui The AUI identifying specific context text to be matched.
	 * @param codingSchemeName The coding scheme container the code to be qualified.
	 * @param qualifierName The qualifier name (e.g. 'HCD').
	 * @param qualifierValue The qualifier value.
	 * @throws SQLException
	 */
	private void qualifyConceptPresentation(String code, String aui, String codingSchemeName, String qualifierName, String qualifierValue) throws SQLException {
		///////////////////////////////////////////////////////////////////////
		// Note: Separate the RRF and LexGrid queries into separate SQL since
		// content for each may be stored in separate databases (multi-db mode).
		// Syntax (and ability) to query across databases varies by database
		// engine.
		///////////////////////////////////////////////////////////////////////
		
		// Locate matching root entry in MRCONSO ...
		PreparedStatement getMRCONSO = umlsConnection2_.prepareStatement(
				"SELECT CODE, STR from MRCONSO WHERE MRCONSO.CODE = ? AND MRCONSO.AUI = ?"
			);
		getMRCONSO.setString(1, code);
		getMRCONSO.setString(2, aui);
		try {
	    	// For each, find matching concept properties (text presentations) and add
			// a property qualifier.  Note: multiple qualifiers can exist per term.
			// Note: Oracle was pretty finicky about how we defined the schema to allow
			// for very large property values.  Unfortunately it also does not allow
			// query over these values (stored as CLOBS) without use of the DBMS_LOB
			// package functions, and there it appeared we could not do better than a
			// 'startsWith' match.  For now, we strike a compromise by allowing the
			// query to be narrower for non-Oracle, but supporting Oracle by iterating
			// through the returned presentations to find a match.
	    	boolean narrowQuery = !sqlModifier_.getDatabaseType().equals("Oracle");
	    	StringBuffer stmt =
	   			new StringBuffer("SELECT propertyId, propertyValue from ")
	   				.append(stc_.getTableName(SQLTableConstants.CONCEPT_PROPERTY))
	   				.append(" WHERE conceptCode = ? AND propertyType = 'presentation'");
	    	if (narrowQuery)
	    		stmt.append(" AND propertyValue = ? ");
	    	
	    	PreparedStatement getPropIds =
	    		sqlConnection_.prepareStatement(stmt.toString());
	
	    	ResultSet mrconso_results = getMRCONSO.executeQuery();
	    	try {
	    		while (mrconso_results.next()) {
	    			getPropIds.setString(1, mrconso_results.getString(1));
	    			String textToMatch = mrconso_results.getString(2);
	    			if (narrowQuery)
	    				getPropIds.setString(2, textToMatch);
	    			
	    			ResultSet propId_results = getPropIds.executeQuery();
	    			try {
	    		    	while (propId_results.next()) {
	    		    		if (narrowQuery || textToMatch.equals(propId_results.getString(2))) {
		    		    		addConceptToConceptPropertyMultiAttributes(
		    		    				codingSchemeName, 
		    		    				code,
		    		    				propId_results.getString(1),
		    		    				"qualifier",
		    		    				qualifierName, 
		    		    				qualifierValue,
		    		    				"");
		    		    		break;
	    		    		}
	    		    	}
	    			} catch (SQLException e) {
	    				// Warn and continue ...
	    				messages_.warn("Unable to register context for code: " + code + " aui: " + aui,
	    					e);
	    			} finally {
	    				propId_results.close();
	    			}
	    		}
	    	} finally {
	    		getPropIds.close();
			}
		} finally {
			getMRCONSO.close();
		}    	
	}

    ////////////////////////////////////////////////////////////////
    // Code for tracking hierarchical associations used in each
	// container, and according to the direction of navigation from
	// parent to child...
    ////////////////////////////////////////////////////////////////

	/**
	 * Build navigable relations from special endpoint nodes '@' and '@@' to
	 * first level nodes in hierarchical associations.
	 * 
	 * @param codingSchemeName
	 *            The LexGrid code system name.
	 * @throws SQLException
	 */
    protected void buildRootNodes(String codingSchemeName) throws SQLException
    {
    	// Calculate for the source relations container ...
    	SABString[] names = getHierRelas();
    	for (int i = 0; i < names.length; i++) {
    		SABString name = names[i];
    		if (isLoaded(name.str, codingSchemeName, SRC_RELATIONS)) {
            	buildRootNode(codingSchemeName, false, new SABString[] {name}, SRC_RELATIONS);
    		}
    	}
    	
    	// Calculate for standard umls relations ...
    	names = getHierRels();
    	for (int i = 0; i < names.length; i++) {
    		SABString name = names[i];
    		if (isLoaded(name.str, codingSchemeName, STD_RELATIONS)) {
            	buildRootNode(codingSchemeName, false, new SABString[] {name}, STD_RELATIONS);
    		}
    	}
    }
    
}