/*
 * Copyright: (c) 2004-2007 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package edu.mayo.informatics.lexgrid.convert.directConversions.UmlsCommon;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.LexGrid.messaging.LgMessageDirectorIF;
import org.LexGrid.util.sql.DBUtility;
import org.LexGrid.util.sql.GenericSQLModifier;
import org.LexGrid.util.sql.lgTables.SQLTableConstants;
import org.LexGrid.util.sql.lgTables.SQLTableUtilities;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import edu.mayo.informatics.lexgrid.convert.utility.Constants;

/**
 * Common Code for RRF/UMLS based loaders.
 * 
 * Here are the table use details for the UMLS converter
 * 
 * Used tables:
 * 
 * MRCONSO.RRF MRREL.RRF MRSAB.RRF MRDEF.RRF MRSTY.RRF MRSAT.RRF MRDOC.RRF
 * 
 * Conditionally referenced: MRHIER.RRF (for contextual links)
 * 
 * Unused tables: AMBIGLUI.RRF AMBIGSUI.RRF MRAUI.RRF MRCOC.RRF MRCOLS.RRF MRCUI.RRF MRCXT.RRF 
 * MRFILES.RRF MRHIST.RRF MRMAP.RRF MRRANK.RRF MRSMAP.RRF MRXNS_ENG.RRF MRXNW_ENG.RRF MRXW_BAQ.RRF
 * MRXW_CZE.RRF MRXW_DAN.RRF MRXW_DUT.RRF MRXW_ENG.RRF MRXW_FIN.RRF MRXW_FRE.RRF MRXW_GER.RRF MRXW_HEB.RRF
 * MRXW_HUN.RRF MRXW_ITA.RRF MRXW_JPN.RRF MRXW_NOR.RRF MRXW_POR.RRF MRXW_RUS.RRF MRXW_SPA.RRF MRXW_SWE.RRF
 * 
 * 
 * There are a couple of unused tables that I probably should be using - but I'm just not loading that info
 * yet - such as the history. I'm not sure about the map tables - I'll need to look into them more at some
 * future point.
 * 
 * 
 * Here is the breakdown for the NCI MetaThesaurus:
 * 
 * Used tables:
 * 
 * MRCONSO.RRF MRREL.RRF MRSAB.RRF MRDEF.RRF MRSTY.RRF
 * 
 * 
 * Unused tables: MRCOC.RRF MRCOLS.RRF MRDOC.RRF MRFILES.RRF MRRANK.RRF
 * 
 * Here, I think that the only one that I'm not using that I maybe should be is the MRCOC file - I'll have to
 * look at this again at some point.
 * 
 * Differences/issues I can remember (though there may have been more):
 * 
 * While loading from the UMLS: For most terminologies, I could get the Language from the LAT column in the
 * MRCONSO file. In Snomed, I had to do a query on the MRSAT table and pull the language out of the ATV
 * column.
 * 
 * In the UMLS - when we load things, we are loading based on terminology concept codes, not CUI's - so to get
 * the preferred text for a code in a terminology (per language) required reading all presentations for a code
 * from a terminology, and then doing a sort taking into account 4 different columns, each with various values
 * that mean preferred in different terminologies
 * 
 * In the NCI MetaThesaurus, since we are loading it using the CUI's as the concept codes, I can just use the
 * IsPref column to determine the preferred text.
 * 
 * 
 * I think that all of the other differences in the UMLS loader versus the NCI MetaThesaurus loader are due to
 * loading based on CUI's instead of concept codes.
 * 
 * @author <A HREF="mailto:armbrust.daniel@mayo.edu">Dan Armbrust</A>
 * @author <A HREF="mailto:erdmann.jesse@mayo.edu">Jesse Erdmann</A>
 * @author <A HREF="mailto:johnson.thomas@mayo.edu">Thomas Johnson</A>
 * @version subversion $Revision: 5997 $ checked in on $Date: 2007-09-28 03:46:34 -0500 (Fri, 28 Sep 2007) $
 */
public class UMLSBaseCode
{
    protected static Logger       log;

    protected Connection          umlsConnection_;
    protected Connection          umlsConnection2_;
    protected Connection          sqlConnection_;

    protected SQLTableUtilities   sqlTableUtility_;
    protected SQLTableConstants   stc_;
    protected GenericSQLModifier  umlsSqlModifier_;
    protected GenericSQLModifier  sqlModifier_;

    protected final int           batchSize                   = Constants.mySqlBatchSize;

    protected Hashtable           isoMap                      = new Hashtable();

    protected HashSet             loadedAssociations_         = new HashSet();
    protected HashSet             supportedCodingSchemes_     = new HashSet();
    protected Association[]       supportedAssociations_      = null;
    protected HashSet             supportedAssociationQualifiers_ = new HashSet();
    protected HashSet             supportedPropertyTypes_     = new HashSet();
    protected HashSet             supportedPropertyQualifiers_= new HashSet();
    protected HashSet             supportedSources_           = new HashSet(); 
    protected HashSet             supportedLanguages_         = new HashSet();
    protected String              defaultLanguage_;

    protected Hashtable           mrconsoRepresentationalMap_ = new Hashtable();
	protected Hashtable sabToCodeSystem_ = new Hashtable();

    protected PreparedStatement   insertIntoCodingScheme;
    protected PreparedStatement   insertIntoCodingSchemeMultiAttributes;
    protected PreparedStatement   insertIntoConceptProperty;
    protected PreparedStatement   insertIntoConceptPropertyMultiAttributes;
    protected PreparedStatement   insertIntoConcepts;
    protected PreparedStatement   insertIntoConceptsMultiAttributes;
    protected PreparedStatement   insertIntoRelations;
    protected PreparedStatement   insertIntoAssociations;
    protected PreparedStatement   insertIntoConceptAssociationsToConcept;
    protected PreparedStatement   insertIntoConceptAssociationsToConceptMultiAttributes;
    protected PreparedStatement   insertIntoCodingSchemeSupportedAttributes;
    protected PreparedStatement   insertIntoRelationMultiAttributes;
    protected PreparedStatement   getCodeForCUI_;
    protected PreparedStatement   getCodeForCUINoAUI_;
    protected PreparedStatement   getSource_;
    protected PreparedStatement   getUMLSRoot_;
    protected PreparedStatement   getAssocInstance_;
    protected PreparedStatement   getAssocQualifier_;

    protected LgMessageDirectorIF messages_;
	private MessageDigest md_ = null;

    public class Association implements Comparable
    {
    	public String name;
    	public String rrfName;
    	public String rrfInverse;
    	public String rrfField;
    	public String rrfSAB;
    	public boolean rrfSourceDirectionalityReversed = false;
    	public String toShortString() {
    		return new StringBuffer(64)
				.append(rrfSAB == null ? "" : (rrfSAB + ':'))
				.append(rrfName)
				.toString();
    	}
    	public String toString() {
    		return new StringBuffer(256)
    			.append("Non-directional association name: ").append(name)
    			.append("\n\tRRF name: ").append(rrfName)
    			.append("\n\tRRF inverse: ").append(rrfInverse)
    			.append("\n\tRRF field: ").append(rrfField)
    			.append("\n\tRRF sab: ").append(rrfSAB)
    			.append("\n\tRRF directionity reversed: ").append(rrfSourceDirectionalityReversed)
    			.toString();
    	}
    	public boolean isForwardMatch(SABString qualName) {
    		return
    			((StringUtils.isBlank(rrfSAB)) || StringUtils.isBlank(qualName.sab) || rrfSAB.equalsIgnoreCase(qualName.sab)) &&
    				(StringUtils.isNotBlank(rrfName) && (rrfName.equalsIgnoreCase(qualName.str)));
    	}
    	public boolean isInverseMatch(SABString qualName) {
			return
				((StringUtils.isBlank(rrfSAB)) || StringUtils.isBlank(qualName.sab) || rrfSAB.equalsIgnoreCase(qualName.sab)) &&
					(StringUtils.isNotBlank(rrfInverse) && (rrfInverse.equalsIgnoreCase(qualName.str)));
    	}
		public int compareTo(Object o) {
			if (!(o instanceof Association))
				return -1;
			int i = 0;
			if (name != null
					&& ((i = name.compareTo(((Association) o).name)) != 0))
				return i;
			if (rrfSAB != null 
					&& ((i = rrfSAB.compareTo(((Association) o).rrfSAB)) != 0))
				return i;
			return 0;
		}
    }

    public class AssociationQualification
    {
    	public String qualifierName;
    	public String qualifierValue;
    	public String codingSchemeName;
    	public String pathToRoot;
    	public String sourceConceptAUI;
    	public String sourceConceptCode;
    }
    
    public class AssociationSorter implements Comparator
    {
        public int compare(Object arg0, Object arg1)
        {
        	// Group by field, so that source-defined and umls-defined
        	// relations are grouped together.  Also, keep items with the
        	// same non-directional name together (potentially differentiated
        	// by SAB).
            Association a = (Association) arg0;
            Association b = (Association) arg1;
            int i = -1 * a.rrfField.compareTo(b.rrfField);
            if (i != 0) return i;
            return
            	new SABString(a.name, a.rrfSAB).compareTo
            		(new SABString(b.name, b.rrfSAB));
        }
    }

    public class CodeHolder
    {
        public String code;
        public String codingScheme;
        public CodeHolder() {
        	super();
        }
        public CodeHolder(String code, String codingScheme) {
        	this();
        	this.code = code;
        	this.codingScheme = codingScheme;
        }
    }

    public class ConceptPresentationSorter implements Comparator
    {
        private boolean metaMode = false;

        public ConceptPresentationSorter(boolean NCIMetaThesaurusMode)
        {
            this.metaMode = NCIMetaThesaurusMode;
        }
        
        public ConceptPresentationSorter()
        {
            this.metaMode = false;
        }

        /*
         * This comparator is used for sorting an array of ConceptPresentations. It groups them by language, (default
         * language first) and inside the language groups, it ranks them by their best likleyhood of being the
         * "preferredPresentation"
         */
        public int compare(Object a, Object b)
        {
            ConceptPresentation aa = (ConceptPresentation) (a);
            ConceptPresentation bb = (ConceptPresentation) (b);

            if (aa.language.equals(defaultLanguage_) && !bb.language.equals(defaultLanguage_))
            {
                // aa is default language, bb is not.
                return -1;
            }
            else if (bb.language.equals(defaultLanguage_) && !aa.language.equals(defaultLanguage_))
            {
                // bb is default language, aa is not.
                return 1;
            }
            else if (!aa.language.equals(bb.language))
            {
                // languages not equal to each other, group by languages alphabetically
                return aa.language.compareTo(bb.language);
            }
            else
            // languages equal each other - now order by which one has the "more preferred" presentation
            {
                // NCI MetaThesaurus RRF files actually have the isPref column set in a useful fashion, so we can just
                // use that.

                if (metaMode && aa.ISPREF.equals("Y") && !bb.ISPREF.equals("Y"))
                {
                    // aa has a preferred flag set on the isPref column.
                    return -1;
                }
                else if (metaMode && bb.ISPREF.equals("Y") && !aa.ISPREF.equals("Y"))
                {
                    // bb has a preferred flag set on the isPref column.
                    return 1;
                }
                else
                {
                    //Not in NCI MetaThesaurus mode, or neither had the flag set.
                    if ((aa.TTY.equals("PT") || aa.TTY.equals("MH")) && (!bb.TTY.equals("PT") || !bb.TTY.equals("MH")))
                    {
                        // aa has a preferred flag set on the TTY column, but bb does not
                        return -1;
                    }
                    else if ((bb.TTY.equals("PT") || bb.TTY.equals("MH"))
                            && (!aa.TTY.equals("PT") || !aa.TTY.equals("MH")))
                    {
                        // bb has a preferred flag set on the tty column, but aa does not.
                        return 1;
                    }
                    else
                    {
                        // both or neither have the TTY column set the way I want, look at the TS and STT columns
                        if ((aa.TS.equals("P") && aa.STT.equals("PF")) && (!bb.TS.equals("P") || !bb.STT.equals("PF")))
                        {
                            // aa is preferred, bb is not
                            return -1;
                        }
                        else if ((bb.TS.equals("P") && bb.STT.equals("PF"))
                                && (!aa.TS.equals("P") || !aa.STT.equals("PF")))
                        {
                            // bb is preferred, aa is not.
                            return 1;
                        }
                        else
                        {
                            // both or neither have both columns set the way I want, what about one column?
                            // favor STT column first
                            if (aa.STT.equals("PF") && !bb.STT.equals("PF"))
                            {
                                // aa preferred, bb not.
                                return -1;
                            }
                            else if (bb.STT.equals("PF") && !aa.STT.equals("PF"))
                            {
                                // bb preferred, aa not
                                return 1;
                            }
                            else if (aa.TS.equals("P") && !bb.TS.equals("P"))
                            {
                                return -1;
                            }
                            else if (bb.TS.equals("P") && !aa.TS.equals("P"))
                            {
                                return 1;
                            }
                            else
                            {
                                // neither is preferred over the other.
                                return 0;
                            }
                        }
                    }
                }
            }
        }
    }

    public class ConceptPresentation
    {
        public String  cui;
        public String  representationForm;
        public String  STT = "";
        public String  TS  = "";
        public String  value;
        public String  TTY = "";
        public String  presentationFormat;
        public String  conceptCode;
        public String  language;
        public String  source;
        public String  ISPREF = "";
        public String  AUI = "";
        public Boolean isPreferred;
    }
    
   protected void makeConnections(String umlsServer, String umlsDriver, String umlsUserName, String umlsPassword,
            String sqlServer, String sqlDriver, String sqlUserName, String sqlPassword, String tablePrefix, boolean enforceIntegrity) throws Exception
    {
        try
        {
            umlsConnection_ = DBUtility.connectToDatabase(umlsServer, umlsDriver, umlsUserName, umlsPassword);
            umlsConnection2_ = DBUtility.connectToDatabase(umlsServer, umlsDriver, umlsUserName, umlsPassword);
            sqlConnection_ = DBUtility.connectToDatabase(sqlServer, sqlDriver, sqlUserName, sqlPassword);
        }
        catch (ClassNotFoundException e)
        {
            messages_.info("The class you specified for your sql driver could not be found on the path.");
        }

        sqlTableUtility_ = new SQLTableUtilities(sqlConnection_, tablePrefix);
        stc_ = sqlTableUtility_.getSQLTableConstants();
        
        umlsSqlModifier_ = new GenericSQLModifier(umlsConnection_);
        sqlModifier_ = new GenericSQLModifier(sqlConnection_);

        messages_.info("Creating tables");
        sqlTableUtility_.createDefaultTables();
        if (enforceIntegrity)
        {
            messages_.info("Creating constraints");
            sqlTableUtility_.createDefaultTableConstraints();
        }
        else
        {
            messages_.info("Removing constraints");
            sqlTableUtility_.dropDefaultTableConstraints();
        }

        initLoadStatements();

    }

    /*
     * enter the mappings from MRCONSO TTY codes to the "nice names"
     */
    protected void initMRCONSOTTYMAP()
    {
        mrconsoRepresentationalMap_.put("FN", "Full Form");
        // TODO - add more of these - see http://www.nlm.nih.gov/research/umls/metab3.html#sb3_0
    }

    protected void addToConceptProperty(String codingSchemeName, String conceptCode, String propertyId,
            String propertyType, String property, String language, String presentationFormat, Boolean isPreferred,
            String degreeOfFidelity, Boolean matchIfNoContext, String representationForm, String propertyValue)
            throws SQLException
    {
        int k = 1;
        insertIntoConceptProperty.setString(k++, codingSchemeName);
        insertIntoConceptProperty.setString(k++, conceptCode);
        insertIntoConceptProperty.setString(k++, propertyId);
        if (stc_.supports2006Model())
        {
            insertIntoConceptProperty.setString(k++, propertyType);
        }
        insertIntoConceptProperty.setString(k++, property);
        insertIntoConceptProperty.setString(k++, language);
        insertIntoConceptProperty.setString(k++, presentationFormat);
        if (!stc_.supports2006Model())
        {
            insertIntoConceptProperty.setString(k++, "");
        }
        DBUtility.setBooleanOnPreparedStatment(insertIntoConceptProperty, k++, isPreferred);
        insertIntoConceptProperty.setString(k++, degreeOfFidelity);
        DBUtility.setBooleanOnPreparedStatment(insertIntoConceptProperty, k++, matchIfNoContext);
        insertIntoConceptProperty.setString(k++, representationForm);
        insertIntoConceptProperty.setString(k++, propertyValue);

        insertIntoConceptProperty.executeUpdate();
    }

    protected void addConceptToConceptPropertyMultiAttributes(String codingSchemeName, String conceptCode,
            String propertyId, String typeName, String attributeValue, String val1, String val2) throws SQLException
    {
        insertIntoConceptPropertyMultiAttributes.setString(1, codingSchemeName);
        insertIntoConceptPropertyMultiAttributes.setString(2, conceptCode);
        insertIntoConceptPropertyMultiAttributes.setString(3, propertyId);
        insertIntoConceptPropertyMultiAttributes.setString(4, typeName);
        insertIntoConceptPropertyMultiAttributes.setString(5, attributeValue);
        if (stc_.supports2006Model())
        {
            insertIntoConceptPropertyMultiAttributes.setString(6, StringUtils.isNotBlank(val1) ? val1 : " ");
            insertIntoConceptPropertyMultiAttributes.setString(7, val2);
        }

        insertIntoConceptPropertyMultiAttributes.executeUpdate();
    }

    protected void addConceptToConceptsMultiAttributes(String codingSchemeName, String conceptCode,
            String attributeName, String attributeValue) throws SQLException
    {

        insertIntoConceptsMultiAttributes.setString(1, codingSchemeName);
        insertIntoConceptsMultiAttributes.setString(2, conceptCode);
        insertIntoConceptsMultiAttributes.setString(3, attributeName);
        insertIntoConceptsMultiAttributes.setString(4, attributeValue);

        insertIntoConceptsMultiAttributes.executeUpdate();
    }

    protected void addAssociationToAssociations(String codingSchemeName, String relationName, String association,
            String forwardName, String reverseName, String inverse, Boolean isNavigable, Boolean isTransitive, Boolean isAntiTransitive,
            Boolean isSymmetric, Boolean isAntiSymmetric, Boolean isReflexive, Boolean isAntiReflexive,
            Boolean isFunctional, Boolean isReverseFunctional, Boolean isTranslationAssociation,
            String targetCodingScheme, String entityDescription) throws SQLException
    {
        insertIntoAssociations.setString(1, codingSchemeName);
        insertIntoAssociations.setString(2, relationName);
        insertIntoAssociations.setString(3, association);
        insertIntoAssociations.setString(4, forwardName);
        insertIntoAssociations.setString(5, reverseName);
        insertIntoAssociations.setString(6, inverse);
        DBUtility.setBooleanOnPreparedStatment(insertIntoAssociations, 7, isNavigable);
        DBUtility.setBooleanOnPreparedStatment(insertIntoAssociations, 8, isTransitive);
        DBUtility.setBooleanOnPreparedStatment(insertIntoAssociations, 9, isAntiTransitive);
        DBUtility.setBooleanOnPreparedStatment(insertIntoAssociations, 10, isSymmetric);
        DBUtility.setBooleanOnPreparedStatment(insertIntoAssociations, 11, isAntiSymmetric);
        DBUtility.setBooleanOnPreparedStatment(insertIntoAssociations, 12, isReflexive);
        DBUtility.setBooleanOnPreparedStatment(insertIntoAssociations, 13, isAntiReflexive);
        DBUtility.setBooleanOnPreparedStatment(insertIntoAssociations, 14, isFunctional);
        DBUtility.setBooleanOnPreparedStatment(insertIntoAssociations, 15, isReverseFunctional);
        DBUtility.setBooleanOnPreparedStatment(insertIntoAssociations, 16, isTranslationAssociation);
        insertIntoAssociations.setString(17, targetCodingScheme);
        insertIntoAssociations.setString(18, entityDescription);

        insertIntoAssociations.execute();
    }

    protected void addConceptAssociationToConcepts(String codingSchemeName, String sourceCodingSchemeName,
            String sourceConceptCode, String relationName, String association, String targetCodingSchemeName,
            String targetConceptCode, String multiAttributesKey, Boolean firstVersion, Boolean lastVersion)
            throws SQLException
    {
        insertIntoConceptAssociationsToConcept.setString(1, codingSchemeName);
        insertIntoConceptAssociationsToConcept.setString(2, relationName);
        insertIntoConceptAssociationsToConcept.setString(3, association);
        insertIntoConceptAssociationsToConcept.setString(4, sourceCodingSchemeName);
        insertIntoConceptAssociationsToConcept.setString(5, sourceConceptCode);
        insertIntoConceptAssociationsToConcept.setString(6, targetCodingSchemeName);
        insertIntoConceptAssociationsToConcept.setString(7, targetConceptCode);
        insertIntoConceptAssociationsToConcept.setString(8, multiAttributesKey);
        DBUtility.setBooleanOnPreparedStatment(insertIntoConceptAssociationsToConcept, 9, firstVersion);
        DBUtility.setBooleanOnPreparedStatment(insertIntoConceptAssociationsToConcept, 10, lastVersion);
        if (stc_.supports2006Model())
        {
            DBUtility.setBooleanOnPreparedStatment(insertIntoConceptAssociationsToConcept, 11, lastVersion);
        }

        insertIntoConceptAssociationsToConcept.execute();
    }
    
    protected void addConceptAssociationQualifierToConceptAssociation(String codingSchemeName,
            String multiAttributesKey, String qualifierName, String qualifierValue) throws SQLException
    {
        insertIntoConceptAssociationsToConceptMultiAttributes.setString(1, codingSchemeName);
        insertIntoConceptAssociationsToConceptMultiAttributes.setString(2, multiAttributesKey);
        if (stc_.supports2006Model())
        {
            insertIntoConceptAssociationsToConceptMultiAttributes.setString(3, qualifierName);
            insertIntoConceptAssociationsToConceptMultiAttributes.setString(4, qualifierValue);
        }
        else
        {
            insertIntoConceptAssociationsToConceptMultiAttributes.setString(3, "qualifier");
            insertIntoConceptAssociationsToConceptMultiAttributes.setString(4, qualifierName);
            insertIntoConceptAssociationsToConceptMultiAttributes.setString(5, "");
            insertIntoConceptAssociationsToConceptMultiAttributes.setString(6, qualifierValue);
        }

        insertIntoConceptAssociationsToConceptMultiAttributes.execute();
    }

    protected void addRelationToRelations(String codingSchemeName, String relationName, Boolean isNative,
            String entityDescription) throws SQLException
    {

        insertIntoRelations.setString(1, codingSchemeName);
        insertIntoRelations.setString(2, relationName);
        DBUtility.setBooleanOnPreparedStatment(insertIntoRelations, 3, isNative);
        insertIntoRelations.setString(4, entityDescription);

        insertIntoRelations.execute();
    }

    protected void addConceptToConcepts(String codingSchemeName, String conceptCode, Boolean firstRelease, Boolean modifiedInRelease,
            Boolean deprecated, Boolean isActive, String conceptStatus, Boolean isAnonymous, String entityDescription)
            throws SQLException
    {
        int k = 1;
        insertIntoConcepts.setString(k++, codingSchemeName);
        insertIntoConcepts.setString(k++, conceptCode);
        DBUtility.setBooleanOnPreparedStatment(insertIntoConcepts, k++, firstRelease);
        if (stc_.supports2006Model())
        {
            DBUtility.setBooleanOnPreparedStatment(insertIntoConcepts, k++, modifiedInRelease);
        }
        DBUtility.setBooleanOnPreparedStatment(insertIntoConcepts, k++, deprecated);
        DBUtility.setBooleanOnPreparedStatment(insertIntoConcepts, k++, isActive);
        insertIntoConcepts.setString(k++, conceptStatus);
        DBUtility.setBooleanOnPreparedStatment(insertIntoConcepts, k++, isAnonymous);
        insertIntoConcepts.setString(k++, entityDescription);

        insertIntoConcepts.execute();
    }

    protected void addAttributeToRelationMultiAttributes(String codingSchemeName, String relationName,
            String attributeName, String attributeValue) throws SQLException
    {
        insertIntoRelationMultiAttributes.setString(1, codingSchemeName);
        insertIntoRelationMultiAttributes.setString(2, relationName);
        insertIntoRelationMultiAttributes.setString(3, attributeName);
        insertIntoRelationMultiAttributes.setString(4, attributeValue);
        if (stc_.supports2006Model())
        {
            insertIntoRelationMultiAttributes.setString(5, "");
            insertIntoRelationMultiAttributes.setString(6, "");
        }

        insertIntoRelationMultiAttributes.execute();
    }
    
    public static Hashtable getIsoMap() throws Exception
    {
        String fileName = "/UMLS_SAB_ISO_Map.txt";
        Hashtable result = new Hashtable();
        BufferedReader in;
        int lineNo = 1;

        in = new BufferedReader(new InputStreamReader(UMLSBaseCode.class.getResourceAsStream(fileName)));
        String line = in.readLine();
        while (line != null)
        {
            if (!line.startsWith("#") && line.length() > 0)
            {
                String[] foo = line.split("=");
                if (foo.length == 2)
                {
                    result.put(foo[0], foo[1]);
                }
                else
                {
                    throw new Exception("Invalid format on line " + lineNo);
                }
            }
            line = in.readLine();
            lineNo++;
        }
        in.close();
        return result;
    }

    /*
     * enter in the mappings of coding schemes to iso codes
     */
    protected void initIsoMap()
    {
        messages_.info("Loading the SAB -> URN map");
        
        // expects a file with one entry per line
        // each line looks like:
        // SAB=urn:oid:x.x.x.x
        try
        {
            isoMap = getIsoMap();
        }
        catch (Exception e)
        {
            messages_.info("Problem trying to read the file containing the SAB -> ISO map");
            log.error("Problem trying to read the file containing the SAB -> ISO map", e);
        }
    }
    
    protected void updateApproxNumberOfConcepts(int count, String codingSchemeName)
    {
        try
        {
            messages_.info("Updating concept count");
            PreparedStatement temp = sqlConnection_.prepareStatement("UPDATE " + stc_.getTableName(SQLTableConstants.CODING_SCHEME) + " SET approxNumConcepts = ? WHERE codingSchemeName = ?");
            temp.setInt(1, count);
            temp.setString(2, codingSchemeName);
            temp.executeUpdate();
            temp.close();
        }
        catch (SQLException e)
        {
            messages_.error("Problem updating the concept count", e);
            log.error("Problem updating the concept count", e);
        }
    } 

    /*
     * make sure it doesn't return null, in case it is missing. This maps either RSABs or VSABS to the iso
     * code.
     */
    protected String getISOString(String UMLSCodingSchemeName) throws SQLException
    {
        String temp = (String) isoMap.get(UMLSCodingSchemeName);
        if (temp == null)
        {
            // maybe it was a VSAB instead of an RSAB.
            String RSAB = mapVSABtoRSAB(UMLSCodingSchemeName);
            temp = (String) isoMap.get(RSAB);
            if (temp == null)
            {
                log.error("Could not get the proper URN for " + UMLSCodingSchemeName);
                messages_.info("Could not get the proper URN for " + UMLSCodingSchemeName);
                temp = "Unknown-" + Math.random();
            }
        }
        return temp;
    }
    
    protected String mapVSABtoRSAB(String vsab) throws SQLException
    {
        PreparedStatement getRSAB = umlsConnection_.prepareStatement("SELECT RSAB FROM MRSAB WHERE VSAB = ?");

        getRSAB.setString(1, vsab);

        ResultSet results = getRSAB.executeQuery();
        String result = "";
        if (results.next())
        {
            result = results.getString("RSAB");
        }

        results.close();
        getRSAB.close();
        return result;
    }

    /*
     * populate the properties attributes of the codingSchemeSuppportedAttributes table
     */
    protected void loadSupportedProperties(String codingSchemeName) throws SQLException
    {
        messages_.info("loading the supported properties");
        String[] temp = (String[]) supportedPropertyTypes_.toArray(new String[supportedPropertyTypes_.size()]);
        for (int i = 0; i < temp.length; i++)
        {
            insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "Property", temp[i], null, null, null, null);
        }
        messages_.info("loaded " + temp.length + " properties");
    }
    
    /*
     * populate the properties attributes of the codingSchemeSuppportedAttributes table
     */
    protected void loadSupportedPropertyQualifiers(String codingSchemeName) throws SQLException
    {
        messages_.info("loading the supported property qualifiers");
        String[] temp = (String[]) supportedPropertyQualifiers_.toArray(new String[supportedPropertyQualifiers_.size()]);
        for (int i = 0; i < temp.length; i++)
        {
            insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "PropertyQualifier", temp[i], null, null, null, null);
        }
        messages_.info("loaded " + temp.length + " property qualifiers");
    }

    /*
     * populate the properties attributes of the codingSchemeSuppportedAttributes table
     */
    protected void loadSupportedAssociationQualifiers(String codingSchemeName) throws SQLException
    {
        messages_.info("loading the supported association qualifiers.");
        String[] temp = (String[]) supportedAssociationQualifiers_.toArray(new String[supportedAssociationQualifiers_.size()]);
        for (int i = 0; i < temp.length; i++)
        {
            insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "AssociationQualifier", temp[i], null, null, null, null);
        }
        messages_.info("loaded " + temp.length + " association qualifiers");
    }

    /*
     * populate the properties attributes of the codingSchemeSuppportedAttributes table
     */
    protected void loadSupportedSources(String codingSchemeName) throws SQLException
    {
        messages_.info("loading the supported sources");
        String[] temp = (String[]) supportedSources_.toArray(new String[supportedSources_.size()]);
        for (int i = 0; i < temp.length; i++)
        {
            insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "Source", temp[i],
                                                      getISOString(temp[i]), null, null, null);

        }
        messages_.info("loaded " + temp.length + " sources");
    }

   /*
     * populate the properties attributes of the codingSchemeSuppportedAttributes table
     */
    protected void loadSupportedLanguages(String codingSchemeName) throws SQLException
    {
        messages_.info("loading the supported languages");
        String[] temp = (String[]) supportedLanguages_.toArray(new String[supportedLanguages_.size()]);
        for (int i = 0; i < temp.length; i++)
        {
            // TODO figure out how to get the URN's and put them in here
            insertIntoCodingSchemeSupportedAttributes(codingSchemeName, "Language", temp[i], null, null, null, null);

        }
        messages_.info("loaded " + temp.length + " languages");
    }

    /*
     * Method to do the insertion into the table
     */
    protected void insertIntoCodingSchemeSupportedAttributes(String codingSchemeName, String supportedAttributeTag,
            String id, String urn, String value, String assemblyRule, String agentRole) throws SQLException
    {
        insertIntoCodingSchemeSupportedAttributes.setString(1, codingSchemeName);
        insertIntoCodingSchemeSupportedAttributes.setString(2, supportedAttributeTag);
        insertIntoCodingSchemeSupportedAttributes.setString(3, id);
        insertIntoCodingSchemeSupportedAttributes.setString(4, urn);
        if (stc_.supports2006Model())
        {
            insertIntoCodingSchemeSupportedAttributes.setString(5, value);
            insertIntoCodingSchemeSupportedAttributes.setString(6, assemblyRule);
            insertIntoCodingSchemeSupportedAttributes.setString(7, agentRole);
        }

        insertIntoCodingSchemeSupportedAttributes.executeUpdate();
    }
    
    /*
     * Create the statements that insert into the tables
     */
    protected void initLoadStatements() throws SQLException
    {
        insertIntoCodingScheme = sqlConnection_.prepareStatement(stc_
                .getInsertStatementSQL(SQLTableConstants.CODING_SCHEME));

        insertIntoCodingSchemeSupportedAttributes = sqlConnection_.prepareStatement(stc_
                .getInsertStatementSQL(SQLTableConstants.CODING_SCHEME_SUPPORTED_ATTRIBUTES));

        insertIntoCodingSchemeMultiAttributes = sqlConnection_.prepareStatement(stc_
                .getInsertStatementSQL(SQLTableConstants.CODING_SCHEME_MULTI_ATTRIBUTES));

        insertIntoConcepts = sqlConnection_.prepareStatement(stc_.getInsertStatementSQL(SQLTableConstants.CONCEPT));

        insertIntoConceptProperty = sqlConnection_.prepareStatement(stc_
                .getInsertStatementSQL(SQLTableConstants.CONCEPT_PROPERTY));

        insertIntoConceptPropertyMultiAttributes = sqlConnection_.prepareStatement(stc_
                .getInsertStatementSQL(SQLTableConstants.CONCEPT_PROPERTY_MULTI_ATTRIBUTES));

        insertIntoRelations = sqlConnection_.prepareStatement(stc_.getInsertStatementSQL(SQLTableConstants.RELATION));

        insertIntoRelationMultiAttributes = sqlConnection_.prepareStatement(stc_
                .getInsertStatementSQL(SQLTableConstants.RELATION_MULTI_ATTRIBUTES));

        insertIntoAssociations = sqlConnection_
                .prepareStatement(stc_.getInsertStatementSQL(SQLTableConstants.ASSOCIATION));

        insertIntoConceptAssociationsToConcept = sqlConnection_.prepareStatement(stc_
                .getInsertStatementSQL(SQLTableConstants.CONCEPT_ASSOCIATION_TO_CONCEPT));

        insertIntoConceptAssociationsToConceptMultiAttributes = sqlConnection_.prepareStatement(stc_
                .getInsertStatementSQL(SQLTableConstants.CONCEPT_ASSOCIATION_TO_C_QUALS));
        
        getCodeForCUI_ = umlsConnection_.prepareStatement("SELECT CODE, SAB FROM MRCONSO WHERE CUI = ? AND AUI = ?");
        
        getCodeForCUINoAUI_ = umlsConnection_.prepareStatement("SELECT Distinct CODE, SAB FROM MRCONSO WHERE CUI = ? AND SAB Like ?");

        getSource_ = umlsConnection_.prepareStatement("SELECT EXPL FROM MRDOC WHERE DOCKEY = ? AND VALUE = ?");
 
        getUMLSRoot_ = umlsConnection_.prepareStatement("SELECT CUI, CODE, SAB FROM MRCONSO WHERE CUI IN (SELECT CUI FROM MRCONSO WHERE SAB = 'SRC' AND TTY = 'RHT')");

        getAssocInstance_ = sqlConnection_.prepareStatement(
        		new StringBuffer(256)
        			.append("SELECT * FROM ").append(stc_.getTableName(SQLTableConstants.CONCEPT_ASSOCIATION_TO_CONCEPT))
					.append(" WHERE codingSchemeName = ? AND relationName = ? AND association = ?")
					.append(" AND sourceCodingSchemeName = ? AND sourceConceptCode = ?")
					.append(" AND targetCodingSchemeName = ? AND targetConceptCode = ?")
					.toString());
        getAssocQualifier_ = sqlConnection_.prepareStatement(
        		new StringBuffer(256)
        			.append("SELECT * FROM ").append(stc_.getTableName(SQLTableConstants.CONCEPT_ASSOCIATION_TO_C_QUALS))
					.append(" WHERE codingSchemeName = ? AND multiAttributesKey = ? AND qualifierName = ? AND qualifierValue = ?")
					.toString());
    }
    
    protected String getSourceVersionString(String fallbackString)
    {
        String result = fallbackString;
        try
        {
            getSource_.setString(1, "RELEASE");
            getSource_.setString(2, "umls.release.name");
            ResultSet results = getSource_.executeQuery();
            
            if (results.next())
            {
                result =  results.getString("EXPL");
            }
            results.close();
        }
        catch (SQLException e)
        {
            // Older versions of the umls did not have the version number in a table.
        }
         
        return result;
    }
    
    protected void closeLoadStatements() throws SQLException
    {
        if (insertIntoCodingScheme != null)
        {
            insertIntoCodingScheme.close();
        }
        if (insertIntoCodingSchemeMultiAttributes != null)
        {
            insertIntoCodingSchemeMultiAttributes.close();
        }
        if (insertIntoCodingSchemeSupportedAttributes != null)
        {
            insertIntoCodingSchemeSupportedAttributes.close();
        }
        if (insertIntoConcepts != null)
        {
            insertIntoConcepts.close();
        }
        if (insertIntoConceptsMultiAttributes != null)
        {
            insertIntoConceptsMultiAttributes.close();
        }
        if (insertIntoAssociations != null)
        {
            insertIntoAssociations.close();
        }
        if (insertIntoConceptAssociationsToConcept != null)
        {
            insertIntoConceptAssociationsToConcept.close();
        }
        if (insertIntoConceptAssociationsToConceptMultiAttributes != null)
        {
            insertIntoConceptAssociationsToConceptMultiAttributes.close();
        }
        if (insertIntoConceptProperty != null)
        {
            insertIntoConceptProperty.close();
        }
        if (insertIntoConceptPropertyMultiAttributes != null)
        {
            insertIntoConceptPropertyMultiAttributes.close();
        }
        if (insertIntoRelationMultiAttributes != null)
        {
            insertIntoRelationMultiAttributes.close();
        }
        if (insertIntoRelations != null)
        {
            insertIntoRelations.close();
        }
        if (getCodeForCUI_ != null)
        {
            getCodeForCUI_.close();
        }
        if (getCodeForCUINoAUI_ != null)
        {
            getCodeForCUINoAUI_.close();
        }
        if (getSource_ != null)
        {
            getSource_.close();
        }
        if (getUMLSRoot_ != null)
        {
        	getUMLSRoot_.close();
        }
        if (getAssocInstance_ != null)
        {
        	getAssocInstance_.close();
        }
        if (getAssocQualifier_ != null)
        {
        	getAssocQualifier_.close();
        }
    }
    
 	/*
     * close the sql connections
     */
    protected void closeConnections() throws SQLException
    {
        closeLoadStatements();
        umlsConnection_.close();
        umlsConnection2_.close();
        sqlConnection_.close();
    }

    ////////////////////////////////////////////////////////////////
    // Code for tracking hierarchical associations used in each
	// container, and according to the direction of navigation from
	// parent to child...
    ////////////////////////////////////////////////////////////////

    /**
     * Inner class defining a string optionally qualified by
     * SAB and providing corresponding equality methods, etc, to
     * support use and comparison in various collections.
     * <p>
     * If the given string is intended to match any SAB, the
     * sab-qualifier is null.
     */
    public class SABString implements Comparable {
		public String str = null;
 		public String sab = null;
 		private int hash = 0;
 		
 		public SABString(String str, String sab) {
			super();
			if (str == null)
				throw new IllegalArgumentException("string cannot be null");
			this.str = str;
			this.sab = sab;
		}
 		public int compareTo(Object o) {
			if (!(o instanceof SABString))
				return -1;
			int i = 0;
			if (str != null
					&& ((i = str.compareTo(((SABString) o).str)) != 0))
				return i;
			if (sab != null 
					&& ((i = sab.compareTo(((SABString) o).sab)) != 0))
				return i;
			return 0;
		}
		public boolean equals(Object obj) {
 			return obj != null
 				&& getClass().equals(obj.getClass())
 				&& hashCode() == obj.hashCode();
 		}
 		public int hashCode() {
 			if (hash == 0) {
 				hash = 31 + str.hashCode();
 				if (sab != null) hash = hash * 31 + sab.hashCode();
 			}
 		    return hash;
 		}
		public String toString() {
			return sab == null ? str : (sab + ':' + str);
		}
 	}
   
	private SABString[] hierRelas_ = null;
	private SABString[] hierRels_ = null;
	private Map schemeToHierNames_ = new HashMap();

    /**
	 * Build relations to an artificial root ('@') or tail ('@@') node
	 * to assist with navigation of one or more associations.
	 * If building a root node, nodes acting as source but not as
	 * target are connected to the '@' node.  If building a tail node,
	 * nodes acting as target but not source are connected to the
	 * '@@' node.  Therefore, '@' becomes the ultimate source node and
	 * '@@' the ultimate target node for navigation.
	 * 
	 * @param codingSchemeName
	 *            The LexGrid code system name being loaded.
	 * @param isMeta
	 *            True if the LexGrid code system is to represent multiple UMLS
	 *            ontologies, acting as a single meta-ontology; false if the
	 *            code system represents a single UMLS ontology.  If true,
	 *            LexGrid concept codes and association references are defined
	 *            in terms of UMLS CUI instead of native codes from the source
	 *            ontology.  In addition, relations for meta distributions are
	 *            interpreted as intra-scheme, whereas relations for
	 *            individual coding systems are allowed to cross coding scheme
	 *            boundaries.
	 * @param dirNames
	 *            The names of UMLS directional relations to be evaluated,
	 *            optionally qualified by SAB.
	 * @param lgRelationContainerName
	 *            The LexGrid relations container defining the associations to be
	 *            evaluated, or null to match any relations for the code
	 *            system.
	 * @throws SQLException
	 */
	protected void buildRootNode(String codingSchemeName, boolean isMeta,
			SABString[] dirNames, String lgRelationContainerName)
			throws SQLException
	{
		// Log status
		StringBuffer nameInfo = new StringBuffer();
		for (int i = 0; i < dirNames.length; i++) {
			if (i > 0)
				nameInfo.append(',');
			nameInfo
				.append(dirNames[i].sab != null ? dirNames[i].sab : "<all sabs>")
				.append(':').append(dirNames[i].str);
		}
	    messages_.info("Adding root relations for container " + (lgRelationContainerName == null ? "<any>" : lgRelationContainerName));
	    messages_.info("Based on RRF name(s) = " + nameInfo.toString());
	
	    // Find the possible root entry or entries introduced by the UMLS.
	    ResultSet rs = getUMLSRoot_.executeQuery();
	    Set qualifiedRoots = new HashSet();
	    try {
	    	while (rs.next())
	    		qualifiedRoots.add(
	    			new SABString(rs.getString(isMeta ? "CUI" : "CODE"), rs.getString("SAB")));
	    } finally {
	    	rs.close();
	    }
	    
		// For each possible root, resolve loaded relationships matching the
		// provided name(s) and direction.
		for (Iterator roots = qualifiedRoots.iterator(); roots.hasNext(); ) {
			
			// Extract the code and SAB from the qualified root entry,
			// and determine the coding scheme used for representation of the
			// SAB within the LexGrid repository. 
			SABString root = (SABString) roots.next();
			String rootCodingSchemeName =
				isMeta ? codingSchemeName : mapSABToCodeSystemName(root.sab);
	
			// Check each directional association name (also provided as a SAB-qualified string) ...
			for (int i = 0; i < dirNames.length; i++) {
				SABString dirName = dirNames[i];
				String codingSchemeForRRFName = dirName.sab != null
					? mapSABToCodeSystemName(dirName.sab) : null;
				
				// Only check associations matching the loaded coding scheme.
				// For meta, assume all associations are included within bounds of
				// the load, since all SABs are brought in.
				if (isMeta ||
						codingSchemeForRRFName == null ||
						codingSchemeForRRFName.equals(codingSchemeName))
				{
					// Check the requested name against all registered associations ...
					for (int j = 0; j < supportedAssociations_.length; j++) {
						Association assoc = supportedAssociations_[j];
						boolean matchFwd = assoc.isForwardMatch(dirName);
						boolean matchRev = assoc.isInverseMatch(dirName);
						if (matchFwd || matchRev) {
							// Try resolving both directions ...
							for (int fwd = 1; fwd >= 0; fwd--) {
								boolean resolveFwd = fwd > 0;
								// The registered LexGrid association name matched a UMLS-defined
								// directional name.  Create the SQL statement to resolve the root nodes.
				    			// If meta and a specific code system was specified for the association,
				    			// we still want to make sure any root identified actually participates
				    			// on behalf of the given scheme.  This could come into play if, for
				    			// example, root concepts A and B both participate in partOf relations,
				    			// and the UMLS SAB for A maps partOf (RELA) to 'PAR' (REL) but the
				    			// SAB for B does not.  When loading a meta distribution, the source
			    				// defining the association is stored as a qualifier.
				    			String assocTable = stc_.getTableName(SQLTableConstants.CONCEPT_ASSOCIATION_TO_CONCEPT);
				    			String assocQualTable = stc_.getTableName(SQLTableConstants.CONCEPT_ASSOCIATION_TO_C_QUALS);
				    			StringBuffer sql = new StringBuffer(256);
				    			sql.append("SELECT * FROM ").append(assocTable);
				    			if (isMeta && codingSchemeForRRFName != null)
				    				sql.append(" INNER JOIN ")
				    					.append(assocQualTable).append(" ON ")
				    					.append(assocTable).append(".multiAttributesKey = ")
				    					.append(assocQualTable).append(".multiAttributesKey");
				    			sql.append(resolveFwd
				    					? " WHERE sourceConceptCode = ? AND sourceCodingSchemeName = ? AND targetCodingSchemeName = ?"
				    					: " WHERE targetConceptCode = ? AND targetCodingSchemeName = ? AND sourceCodingSchemeName = ?");
				    			sql.append(" AND association = ?");
				    			if (lgRelationContainerName != null)
				    				sql.append(" AND relationName = ?");
				    			if (isMeta && dirName.sab != null)
				    				sql.append(" AND ")
				    					.append(assocQualTable).append(".qualifierName = '").append(dirName.sab).append("' AND ")
										.append(assocQualTable).append(".qualifierValue = 'Source'");
				    			
				    			// Prepare the statement and assign parameters ...
				    			PreparedStatement getRootRelations =
				    				sqlConnection_.prepareStatement(sqlModifier_.modifySQL(sql.toString()));
				    			int param = 0;
				    			getRootRelations.setString(++param, root.str);
				    			getRootRelations.setString(++param, rootCodingSchemeName);
				    			getRootRelations.setString(++param, codingSchemeName);
				    			getRootRelations.setString(++param, assoc.name);
				    			if (lgRelationContainerName != null)
				        			getRootRelations.setString(++param, lgRelationContainerName );
				
				    			// The query will resolve existing LexGrid associations from the
				    			// UMLS-designated root to corresponding top nodes.
				    			// For each root/topNode combination add a counterpart substituting
				    			// '@' or '@@' in place of the UMLS root.
				    			try {
				    				rs = getRootRelations.executeQuery();
				    				while (rs.next()) {
				    					String newRel = rs.getString("relationName");
				    					String newAssoc = rs.getString("association");
				    					
				    					String newSrc = resolveFwd ? "@" : rs.getString("sourceConceptCode");
				    					String newTgt = resolveFwd ? rs.getString("targetConceptCode") : "@@";
				    					
				    		        	// Try optimistic insert.  On error, check for duplicate and reuse
				    		        	// multiAttributesKey if present.  If not, throw the original err.
				    			        String multiKey = null;
				    		        	try {
			    							multiKey = isMeta
			    								? generateUniqueKey(new String[] {
			    									codingSchemeName, codingSchemeName, 
			    									newSrc, newRel, newAssoc, 
			    									codingSchemeName, newTgt
			    									})
			    								: null;
				    				        addConceptAssociationToConcepts(
					    			        	codingSchemeName, codingSchemeName,
					    			        	newSrc, newRel, newAssoc,
					    			        	codingSchemeName,
					    			        	newTgt, multiKey,
					    			        	null, null);		    			        
				    		        	} catch (SQLException sqle) {
				    			        	// Check for existing association (same link may be defined
				    						// for multiple sources in meta, etc. If found, extract the
				    		        		// existing multi-attribute key for reuse to add qualifiers.
				    						ResultSet rs2 = null;
				    						getAssocInstance_.setString(1, codingSchemeName);
				    						getAssocInstance_.setString(2, newRel);
				    						getAssocInstance_.setString(3, newAssoc);
				    						getAssocInstance_.setString(4, codingSchemeName);
				    						getAssocInstance_.setString(5, newSrc);
				    						getAssocInstance_.setString(6, codingSchemeName);
				    						getAssocInstance_.setString(7, newTgt);
				    						multiKey = null;
				    						try {
				    							rs2 = getAssocInstance_.executeQuery();
				    							if (rs2.next()) {
				    								multiKey = rs2.getString("multiAttributesKey");
				    							} else {
				    								throw sqle;
				    							}
				    						} catch (Exception e) {
				    		        			log.warn("Association instance not inserted.", sqle);
				    						} finally {
				    							if (rs2 != null)
				    								rs2.close();
				    						}
				    					}
				    		            // If inserting for the meta, ensure registration of source
			    		        		// qualifier.  Here we have no alternate path to follow
				    		            // if an error occurs and the qualifier does or does not exist.
				    		            // The load will not be terminated on failure to load a
				    		            // qualifier.  On error, log a warning and continue.
				    		        	if (isMeta && multiKey != null && dirName.sab != null)
					    	        		try {
					    	        			addConceptAssociationQualifierToConceptAssociation(
					    	        				codingSchemeName, multiKey, dirName.sab, "Source");
					    	        		} catch (SQLException sqle) {
					    						ResultSet rs2 = null;
					    						getAssocQualifier_.setString(1, codingSchemeName);
					    						getAssocQualifier_.setString(2, multiKey);
					    						getAssocQualifier_.setString(3, dirName.sab);
					    						getAssocQualifier_.setString(4, "Source");
					    						try {
					    							rs2 = getAssocQualifier_.executeQuery();
					    							if (!rs2.next())
					    								throw sqle;
					    						} catch (Exception e) {
					    		        			log.warn("Association qualifier not inserted.", sqle);
					    						} finally {
					    							if (rs2 != null)
					    								rs2.close();
					    						}
					    	        		}
				    				}
				    			} finally {
				    				if (rs != null)
				    					rs.close();
				    				getRootRelations.close();
				    			}
							}
		    			}
					}
				}
			}
		}
		messages_.info("Finished adding root relations");
	}

	/**
	 * Return the detailed name of the given RRF name;
	 * null if not available.
	 * @param rrfName The UMLS relation name.
	 * @param rrfField The UMLS field (e.g. 'RELA' or 'REL').
	 * @return String
	 */
	protected String getExpandedForm(String rrfName, String rrfField) throws SQLException {
		String result = null;
		if (!StringUtils.isNotBlank(rrfName)) {
			PreparedStatement getInverse = umlsConnection_.prepareStatement("SELECT EXPL FROM MRDOC WHERE DOCKEY = ? AND value = ? and TYPE = ?");
			try {
				getInverse.setString(1, rrfField);
				getInverse.setString(2, rrfName);
				getInverse.setString(3, "expanded_form");
	
				ResultSet results = getInverse.executeQuery();
				if (results.next())
					result = results.getString("EXPL");
				results.close();
			} finally {
				getInverse.close();
			}
		}
		return result;
	}

	/**
	 * Relation names in the RRF (RELA field) for hierarchical associations
	 * specific to the source distribution.
	 * @return String[]
	 * @throws SQLException If an error occurs resolving from the source.
	 */
    protected SABString[] getHierRelas() throws SQLException {
		if (hierRelas_ == null) {
			// Note: Always add isa & inverse_isa; sometimes not mapped by RELA to
			// the standard parent/child (PAR/CHD) RELs (e.g. MeSH maps to 'RN/RB').
			Set names = new HashSet();
			names.addAll(Arrays.asList(getRelationSABs(new String[] {"isa", "inverse_isa"}, "RELA")));
			names.addAll(Arrays.asList(getRelaNames(getHierRels())));
			hierRelas_ = (SABString[]) names.toArray(new SABString[names.size()]);
		}
		return hierRelas_;
	}

	/**
	 * Standard names (REL field in the RRF) for hierarchical associations
	 * in the UMLS.
	 * @return String[]
	 * @throws SQLException If an error occurs resolving from the source.
	 */
    protected SABString[] getHierRels() throws SQLException {
		if (hierRels_ == null) {
			// PAR = hasParent, CHD = hasChild
			// Note: not including RB/RN (relation broader/narrower) since
			// used in contexts that do not always imply subsumption.
			hierRels_ = getRelationSABs(new String[] {"CHD", "PAR"}, "REL");
		}
		return hierRels_;
	}

	/**
	 * Returns the names of the hierarchical associations in use
	 * for the LexGrid code system of the given name.
	 * @param codeSystemName
	 * @return String[]
	 * @throws SQLException If an error occurs resolving from the source.
	 */
	protected String[] getHierAssocNames(String codeSystemName) throws SQLException {
		String[] names;
		if ((names = (String[]) schemeToHierNames_.get(codeSystemName)) == null) {
			Set matches = new HashSet(); 
			String sab = mapCodeSystemNameToSAB(codeSystemName);
			SABString[][] allNames = new SABString[][] {
					getHierRels(), getHierRelas()
					};
			for (int i = 0; i < allNames.length; i++) {
				SABString[] candidates = allNames[i];
				for (int j = 0; j < candidates.length; j++) {
					SABString candidate = candidates[j];
					if (candidate.sab == null || candidate.sab.equalsIgnoreCase(sab))
						matches.add(candidate.str);
				}
			}
			names = (String[]) matches.toArray(new String[matches.size()]);
		}
		return names;
	}
	
	/**
	 * Resolve source-assigned relation names corresponding to the given
	 * array of standard UMLS relation names (e.g. resolve 'part_of'
	 * for 'PAR').
	 * @param relNames SABString[]
	 * @return String[]
	 * @throws SQLException If an error occurs resolving from the source.
	 */
	protected SABString[] getRelaNames(SABString[] relNames) throws SQLException {
		GenericSQLModifier gsm = null;
		try {
			gsm = new GenericSQLModifier(umlsConnection2_);
		} catch (Exception e) {
			throw new SQLException(e.getMessage());
		}
		
		// Locate matching RELA entries in MRREL file ...
		Collection relas = new HashSet();
		for (int i = 0; i < relNames.length; i++) {
			SABString relName = relNames[i];
			StringBuffer sql = new StringBuffer()
				.append("SELECT DISTINCT RELA from MRREL WHERE REL = '")
				.append(relName.str)
				.append('\'');
			if (relName.sab != null) sql
				.append(" AND SAB = '")
				.append(relName.sab)
				.append('\'');
			PreparedStatement getRELA = umlsConnection2_.prepareStatement(gsm.modifySQL(sql.toString()));
			try {
				ResultSet rs = getRELA.executeQuery();
				String s;
				while (rs.next())
					if (StringUtils.isNotBlank(s = rs.getString(1)))
						relas.add(mapRela(s));
				rs.close();
			} finally {
				getRELA.close();
			}
		}

		// Resolve corresponding SABs for each RELA
		Collection sabStrings = new ArrayList();
		for (Iterator relaIt = relas.iterator(); relaIt.hasNext(); ) {
			String relaName = (String) relaIt.next();
			StringBuffer sql = new StringBuffer()
				.append("SELECT DISTINCT SAB from MRREL WHERE RELA = '")
				.append(relaName)
				.append('\'');
			PreparedStatement getSABs = umlsConnection2_.prepareStatement(gsm.modifySQL(sql.toString()));
			try {
				ResultSet rs = getSABs.executeQuery();
				try {
					while (rs.next())
						sabStrings.add(new SABString(relaName, rs.getString(1)));
				} finally {
					rs.close();
				}
			} finally {
				getSABs.close();
			}
		}
		return (SABString[]) sabStrings.toArray(new SABString[sabStrings.size()]);
	}

	/**
	 * Resolve UMLS source abbreviations corresponding to the given
	 * array of source-defined relation names (e.g. resolved from MRREL).
	 * @param rrfNames String[]
	 * @param rrfField 'RELA' or 'REL'.
	 * @return SABString[] Array of corresponding SAB-qualified relation names.
	 * @throws SQLException If an error occurs resolving from the source.
	 */
	protected SABString[] getRelationSABs(String[] rrfNames, String rrfField) throws SQLException {
		GenericSQLModifier gsm = null;
		try {
			gsm = new GenericSQLModifier(umlsConnection2_);
		} catch (Exception e) {
			throw new SQLException(e.getMessage());
		}
		
		// Locate matching RELA entries in MRREL file ...
		Collection sabStrings = new HashSet();
		for (int i = 0; i < rrfNames.length; i++) {
			String relaName = rrfNames[i];
			StringBuffer sql = new StringBuffer()
				.append("SELECT DISTINCT SAB from MRREL WHERE ")
				.append(rrfField)
				.append(" = '")
				.append(relaName)
				.append('\'');
			PreparedStatement getSABs = umlsConnection2_.prepareStatement(gsm.modifySQL(sql.toString()));
			try {
				ResultSet rs = getSABs.executeQuery();
				String s;
				while (rs.next())
					if (StringUtils.isNotBlank(s = rs.getString(1)))
						sabStrings.add(new SABString(relaName, s));
				rs.close();
			} finally {
				getSABs.close();
			}
		}
		return (SABString[]) sabStrings.toArray(new SABString[sabStrings.size()]);
	}

	/**
	 * Return the inverse name of the given RRF-based relation;
	 * null if not available.
	 * @param rrfName The UMLS relation name.
	 * @param rrfField 'RELA' or 'REL'.
	 * @return String
	 */
	protected String getRelationInverseName(String rrfName, String rrfField) throws SQLException {
		String result = null;
	    if (StringUtils.isNotBlank(rrfName)) {
	    	PreparedStatement getInverse = umlsConnection_
	            .prepareStatement("SELECT EXPL FROM MRDOC WHERE DOCKEY = ? AND value = ? and TYPE = ?");
			try {
				getInverse.setString(1, rrfField);
				getInverse.setString(2, rrfName);
				getInverse.setString(3, rrfField.toLowerCase() + "_inverse");
	
				ResultSet results = getInverse.executeQuery();
				if (results.next())
				    result = results.getString("EXPL");	
				results.close();
			} finally {
				getInverse.close();
			}
	    }
	    return result;
	}

	/**
	 * Indicates whether the given association is used to represent hierarchical
	 * relationships in the source being processed.
	 * 
	 * @param assoc
	 *            Association object to check; not null.
	 * @param codingSchemeName
	 *            Name of the LexGrid coding scheme containing the association;
	 *            null to match any coding scheme.
	 * @return true if representing hierarchical relations; false otherwise.
	 * @throws SQLException
	 *             If an error occurs resolving from the source.
	 */
	protected boolean isHierarchicalAssociation(Association assoc, String codingSchemeName) throws SQLException {
		return isHierarchicalName(assoc.rrfName, codingSchemeName)
			|| isHierarchicalName(assoc.rrfInverse, codingSchemeName);
	}
	
	/**
	 * Indicates whether the given name is used to represent hierarchical
	 * relationships in the source being processed.
	 * 
	 * @param dirName
	 *            Directional name of the association to check.
	 * @param codeSystemName
	 *            Name of the LexGrid coding scheme containing the association;
	 *            null to match any coding scheme.
	 * @return true if the name represents hierarchical definitions; false if not.
	 * @throws SQLException
	 *             If an error occurs resolving from the source.
	 */
	protected boolean isHierarchicalName(String name, String codeSystemName) throws SQLException {
		if (name == null)
			return false;
		SABString[][] allNames = new SABString[][] { getHierRels(), getHierRelas() };
		
		String sab = codeSystemName != null ? mapCodeSystemNameToSAB(codeSystemName) : null;
		for (int i = 0; i < allNames.length; i++) {
			SABString[] candidates = allNames[i];
			for (int j = 0; j < candidates.length; j++) {
				SABString candidate = candidates[j];
				if (name.equalsIgnoreCase(candidate.str)
						&& (sab == null || candidate.sab == null || candidate.sab.equalsIgnoreCase(sab)))
					return true;
			}
		}
		return false;
	}
	
	/**
     * Indicates if the given association matches those currently loaded
     * for the given coding scheme and relations container.
     * @param assoc
     * @param codingSchemeName
     * @param relationsContainerName
     * @return true if loaded; false if not.
	 * @throws SQLException
     */
    protected boolean isLoaded(Association assoc, String codingSchemeName, String relationsContainerName) throws SQLException {
    	// Note: Do not include sab on this check; evaluating at LexGrid level.
    	return isLoaded(assoc.name, codingSchemeName, relationsContainerName);
    }
    
	/**
     * Indicates if the given non-directional association name matches those
     * currently loaded for the given coding scheme and relations container.
     * @param name
     * @param codingSchemeName
     * @param relationsContainerName
     * @return true if loaded; false if not.
	 * @throws SQLException
     */
    protected boolean isLoaded(String name, String codingSchemeName, String relationsContainerName) throws SQLException {
    	// Note: markLoaded() key must match ...
    	return loadedAssociations_.contains(
    		new StringBuffer(128)
    			.append(codingSchemeName)
    			.append(relationsContainerName)
    			.append(name)
    			.toString());
    }
    
	/**
     * Sets an indication that the given association has been loaded
     * for the given coding scheme and relations container.
     * @param assoc
     * @param codingSchemeName
     * @param relationsContainerName
	 * @throws SQLException
     */
    protected void markLoaded(Association assoc, String codingSchemeName, String relationsContainerName) throws SQLException {
    	markLoaded(assoc.name, codingSchemeName, relationsContainerName);
    }
    
	/**
     * Sets an indication that the given association has been loaded
     * for the given coding scheme and relations container.
     * @param name
     * @param codingSchemeName
     * @param relationsContainerName
     * @return true if loaded; false if not.
	 * @throws SQLException
     */
    protected void markLoaded(String name, String codingSchemeName, String relationsContainerName) throws SQLException {
    	// Note: isLoaded() key must match ...
    	loadedAssociations_.add(
    		new StringBuffer(128)
    			.append(codingSchemeName)
    			.append(relationsContainerName)
    			.append(name)
    			.toString());
    }
    
	/**
     * Map from the UMLS source abbreviation (SAB) to the coding
     * scheme name used for identification in the LexGrid model.
     * @param sab
     * @return String
     * @throws SQLException
     */
    protected String mapSABToCodeSystemName(String sab) throws SQLException {
		String result = (String) sabToCodeSystem_.get(sab);
		if (result == null) {
			PreparedStatement getCodingSchemeInfo = umlsConnection_
					.prepareStatement("SELECT SSN FROM MRSAB WHERE RSAB = ?");
			try {
				getCodingSchemeInfo.setString(1, sab);
				ResultSet results = getCodingSchemeInfo.executeQuery();
				if (results.next()) {
					result = results.getString("SSN");
					sabToCodeSystem_.put(sab, result);
				}
				results.close();
			} finally {
				getCodingSchemeInfo.close();
			}
		}
		return result;
    }

	/**
	 * Map from the coding scheme name used for identification in the LexGrid
	 * model to the UMLS source abbreviation (SAB).
	 * 
	 * @param codeSystemName
	 * @return String
	 * @throws SQLException
	 */
    protected String mapCodeSystemNameToSAB(String codeSystemName) throws SQLException {
    	String result = "";
		PreparedStatement getCodingSchemeInfo = umlsConnection_
				.prepareStatement("SELECT RSAB FROM MRSAB WHERE SSN = ?");
		try {
			getCodingSchemeInfo.setString(1, codeSystemName);
			ResultSet results = getCodingSchemeInfo.executeQuery();
			if (results.next()) {
				result = results.getString("RSAB");
			}
			results.close();
		} finally {
			getCodingSchemeInfo.close();
		}
		return result;
	}

    /**
	 * Map the provided information to a new Association and add, if new,
	 * add to the given hashtable.
	 * 
	 * @param rrfName
	 *            The name of the relationship as provided by the UMLS.
	 * @param rrfSAB
	 *            The source abbreviation of the UMLS code system.
	 * @param rrfDirFlag
	 *            The UMLS directional flag. Y indicates that this is the
	 *            direction of the RELA relationship in its source; N indicates
	 *            that it is not; otherwise indicates that it is not important
	 *            or has not yet been determined.  (If blank RELA, we interpret
	 *            as 'N', based on empirical review of meta files).
	 * @param rrfField
	 *            Indicates whether the association name is defined as a
	 *            standard UMLS relation ('REL') or source-provided label
	 *            ('RELA').
	 * @param map
	 *            A Map of Associations, keyed by SAB-qualified
	 *            non-directional name.
	 * @return The mapped Association.
	 * @throws SQLException
	 */
    protected Association mapSupportedAssociationsHelper(
    		String rrfName, String rrfSAB, String rrfDirFlag, String rrfField,
    		Map map) throws SQLException
	{
		// Check for an existing Association matching the given info ...
		boolean isRela = "RELA".equalsIgnoreCase(rrfField);
		Association assoc = (Association) map.get(new SABString(isRela ? mapRela(rrfName) : rrfName, rrfSAB));
		if (assoc != null)
			return assoc;
		
		// Also check reverse direction ...
		String reverseName = getRelationInverseName(rrfName, rrfField);
		if (reverseName.equals(rrfName))
			reverseName = "";
		if (StringUtils.isNotBlank(reverseName))
			assoc = (Association) map.get(new SABString(isRela ? mapRela(reverseName) : reverseName, rrfSAB));
		if (assoc != null)
			return assoc;

		// Not found, define and register now ...
		assoc = new Association();

		boolean isReversed = isRela && !("Y".equalsIgnoreCase(rrfDirFlag));
		assoc.rrfName = rrfName;
		assoc.rrfInverse = reverseName;
		assoc.rrfField = rrfField;
		assoc.rrfSAB = rrfSAB;
		assoc.rrfSourceDirectionalityReversed = isReversed;
		assoc.name = isRela ? mapRela(rrfName) : rrfName;

		map.put(new SABString(assoc.name, rrfSAB), assoc);
		return assoc;
	}

	/**
	 * Maps from the given UMLS-provided relationship name to a pre-designated
	 * LexGrid association name (e.g. map from 'inverse_isa' to 'hasSubtype'), if
	 * available. Otherwise, return the rela unmodified.
	 * 
	 * @param rrfName
	 * @return The mapped value.
	 */
	protected String mapRela(String rrfName) {
		return
			("inverse_isa".equalsIgnoreCase(rrfName))
				? "hasSubtype" : rrfName;
	}
	
	/**
	 * Converts the given string array to a single string, separating 
	 * each value by comma and embedding quotes before and after each
	 * value. 
	 * @param strings
	 * @return String
	 */
	protected String toCommaDelimitedWithQuotes(String[] strings) {
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < strings.length; i++) {
			if (i > 0)
				sb.append(',');
			sb.append('\'').append(strings[i]).append('\'');
		}
		return sb.toString();
	}

	/**
	 * Returns a message digest to be used for fixed-length key generation.
s	 */
	protected MessageDigest getSHA1() {
		if (md_ == null)
			try {
				md_ = MessageDigest.getInstance("SHA1");
			} catch (NoSuchAlgorithmException e) {
			}
		return md_;
	}
	
	/**
	 * Generates a unique but reproducible key based on the
	 * given base of string values.  
	 * @param components
	 * @return A unique string based on 20-byte output from
	 * a SHA-1 message digest.
	 * @throws SQLException
	 */
	protected String generateUniqueKey(String[] basis) {
		MessageDigest md = getSHA1();
		md.reset();
		for (int i = 0; i < basis.length; i++)
			if (basis[i] != null)
				md.update(basis[i].getBytes());
		byte[] bytes = md.digest();
		return String.valueOf(Hex.encodeHex(bytes));
	}
    
	/**
	 * Generates a unique but random key, using the given
	 * integer as seed.
	 * @param seed
	 * @return String
	 */
	protected String generateUniqueKey(int seed) {
		return System.currentTimeMillis() + "" + get5DigitRandom() + "" + seed;
	}

	protected String get5DigitRandom() {
		String result = "";
		while (result.length() != 5) {
			result = "" + (int) Math.floor((Math.random() * 100000));
		}
		return result;
	}
    
}