/*
 * Copyright: (c) 2004-2007 Mayo Foundation for Medical Education and
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 *
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * 		http://www.eclipse.org/legal/epl-v10.html
 *
 */

 /*
 * Modified by Kim Ong, NGIT for supporting NCI Bioportal
 */

package gov.nih.nci.lexbig.ext;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.LexGrid.LexBIG.DataModel.Collections.AssociatedConceptList;
import org.LexGrid.LexBIG.DataModel.Collections.AssociationList;
import org.LexGrid.LexBIG.DataModel.Collections.NameAndValueList;
import org.LexGrid.LexBIG.DataModel.Collections.ResolvedConceptReferenceList;
import org.LexGrid.LexBIG.DataModel.Core.AssociatedConcept;
import org.LexGrid.LexBIG.DataModel.Core.Association;
import org.LexGrid.LexBIG.DataModel.Core.CodingSchemeSummary;
import org.LexGrid.LexBIG.DataModel.Core.CodingSchemeVersionOrTag;
import org.LexGrid.LexBIG.DataModel.Core.ResolvedConceptReference;
import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.Extensions.Generic.LexBIGServiceConvenienceMethods;
import org.LexGrid.LexBIG.Impl.LexBIGServiceImpl;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeGraph;
import org.LexGrid.LexBIG.LexBIGService.LexBIGService;
import org.LexGrid.LexBIG.Utility.Constructors;
import org.LexGrid.LexBIG.Utility.ConvenienceMethods;
import org.LexGrid.commonTypes.EntityDescription;
//import org.apache.commons.lang.StringUtils;

import org.LexGrid.LexBIG.DataModel.InterfaceElements.CodingSchemeRendering;
import org.LexGrid.LexBIG.DataModel.Collections.CodingSchemeRenderingList;
import java.util.*;
import java.io.*;

/**
 * Example showing how to determine and display an unsorted list of root and
 * subsumed nodes, up to a specified depth, for hierarchical relationships.
 * It is primarily geared towards the NCI Metathesaurus source.
 *
 * Hierarchies are navigated primarily through the UMLS-defined PAR (has parent)
 * and CHD (has child) relationships.
 *
 * This program accepts two parameters.  The first indicates the depth
 * to display hierarchical relations.  If 0, only the root nodes are displayed.
 * If 1, nodes immediately subsumed by the root are also displayed, etc.
 * If < 0, a default depth of 0 is assumed.
 *
 * The second parameter must provide the source abbreviation (SAB) of the
 * Metathesaurus source to be evaluated (e.g. ICD9CM, MDR, SNOMEDCT).
 */
public class NavigationUtils {
	final static int DEFAULT_DEPTH = 0;
	final static String[] BROADENING_HIER_ASSOC = new String[] {"CHD", "hasSubtype"};
	final static String[] NARROWING_HIER_ASSOC = new String[] {"PAR", "isa", "branch_of", "part_of", "tributary_of"};

	private static Map<String, String> assocFwdNames_ = new HashMap<String, String>();
	private static Map<String, String> assocRevNames_ = new HashMap<String, String>();

    private static PrintWriter pw = null;

	public NavigationUtils() {
		super();
	}

	public static ResolvedConceptReferenceList getRootNodes(
		    LexBIGService lbSvc,
			String sab,
			String scheme,
			CodingSchemeVersionOrTag csvt) throws LBException
	{
		LexBIGServiceConvenienceMethods lbscm = (LexBIGServiceConvenienceMethods) lbSvc.getGenericExtension("LexBIGServiceConvenienceMethods");
        ResolvedConceptReferenceList list = new ResolvedConceptReferenceList();
		// Find the root nodes of the hierarchy (attached to '@' or '@@' nodes).
		String[] assocNames = BROADENING_HIER_ASSOC;
		for (int i = 0; i < assocNames.length; i++)
		{
			try {
				ResolvedConceptReferenceList roots = getTopNodes(lbSvc, scheme, csvt, null, assocNames[i], sab);
				if (roots != null && roots.getResolvedConceptReferenceCount() > 0)
				{
					for (Iterator<ResolvedConceptReference> nodes = roots.iterateResolvedConceptReference(); nodes.hasNext(); )
					{
						list.addResolvedConceptReference(nodes.next());
					}

				}
			} catch (LBException lbe) {};
		}

		assocNames = NARROWING_HIER_ASSOC;
		for (int i = 0; i < assocNames.length; i++)
		{
			try {
				ResolvedConceptReferenceList roots = getEndNodes(lbSvc, scheme, csvt, null, assocNames[i], sab);
				if (roots != null && roots.getResolvedConceptReferenceCount() > 0)
				{
					for (Iterator<ResolvedConceptReference> nodes = roots.iterateResolvedConceptReference(); nodes.hasNext(); )
					{
						list.addResolvedConceptReference(nodes.next());
					}
				}
			} catch (LBException lbe) {}
		}
		return list;
	}


	public static void printRootNodes(ResolvedConceptReferenceList rootNodes)
	{
		if (rootNodes == null || rootNodes.getResolvedConceptReferenceCount() == 0) return;
		for (Iterator<ResolvedConceptReference> nodes = rootNodes.iterateResolvedConceptReference(); nodes.hasNext(); )
		{
			ResolvedConceptReference ref = nodes.next();
			String code = ref.getConceptCode();
			String name = ref.getEntityDescription().getContent();
			displayMessage("Root: " + code + "(" + name + ")");
		}
	}


	protected static void printNodes(HashMap hmap)
	{
		Set keyset = hmap.keySet();
		Iterator it = keyset.iterator();
		while (it.hasNext())
		{
			String asso = (String) it.next();
			ResolvedConceptReferenceList rcrl = (ResolvedConceptReferenceList) hmap.get(asso);
			if (rcrl.getResolvedConceptReferenceCount() == 0) return;
			for (Iterator<ResolvedConceptReference> nodes = rcrl.iterateResolvedConceptReference(); nodes.hasNext(); )
			{
				ResolvedConceptReference ref = nodes.next();
				String code = ref.getConceptCode();
				String name = ref.getEntityDescription().getContent();
				displayMessage("\t" + asso + ": " + code + "(" + name + ")");
			}
		}
	}


	public static void printNodes(PrintWriter pw, HashMap hmap)
	{
		Set keyset = hmap.keySet();
		Iterator it = keyset.iterator();
		while (it.hasNext())
		{
			String asso = (String) it.next();
			ResolvedConceptReferenceList rcrl = (ResolvedConceptReferenceList) hmap.get(asso);
			if (rcrl.getResolvedConceptReferenceCount() == 0) return;
			for (Iterator<ResolvedConceptReference> nodes = rcrl.iterateResolvedConceptReference(); nodes.hasNext(); )
			{
				ResolvedConceptReference ref = nodes.next();
				String code = ref.getConceptCode();
				String name = ref.getEntityDescription().getContent();
				displayMessage(pw, "\t" + asso + ": " + code + "(" + name + ")");
			}
		}
	}

	public static HashMap getChildNodes(
		    LexBIGService lbSvc,
			String scheme,
			String sab,
			CodingSchemeVersionOrTag csvt,
			String code)
	{
		// HashMap key: assoc, value ResolvedConceptReferenceList
		return getAssociatedNodes(
			lbSvc,
			scheme,
			sab,
			csvt,
			code,
			true);
	}

	public static HashMap getParentNodes(
		    LexBIGService lbSvc,
			String scheme,
			String sab,
			CodingSchemeVersionOrTag csvt,
			String code)
	{
		return getAssociatedNodes(
			lbSvc,
			scheme,
			sab,
			csvt,
			code,
			false);
	}

	protected static HashMap getAssociatedNodes(
		    LexBIGService lbSvc,
			String scheme,
			String sab,
			CodingSchemeVersionOrTag csvt,
			String code,
			boolean navigateForward)

	{
		ResolvedConceptReferenceList rcrl = null;
		LexBIGServiceConvenienceMethods lbscm = null;
		HashMap hmap = new HashMap();
		try {
			lbscm = (LexBIGServiceConvenienceMethods) lbSvc.getGenericExtension("LexBIGServiceConvenienceMethods");

			String[] assocNames = BROADENING_HIER_ASSOC;
			for (int i = 0; i < assocNames.length; i++)
			{
				String assocToCheck = assocNames[i];
				Association subsume = navigateForward
					? getChildrenBySource(lbSvc, lbscm, sab, code, null, assocToCheck, scheme, csvt, false)
					: getParentsBySource(lbSvc, lbscm, sab, code, null, assocToCheck, scheme, csvt, false);

				if (subsume != null) {
					AssociatedConceptList subsumedNodes = subsume.getAssociatedConcepts();
					if (subsumedNodes.getAssociatedConceptCount() > 0)
					{
						if (hmap.get(assocToCheck) != null)
						{
							ResolvedConceptReferenceList list = (ResolvedConceptReferenceList) hmap.get(assocToCheck);
							for (Iterator<AssociatedConcept> subsumed = subsumedNodes.iterateAssociatedConcept(); subsumed.hasNext(); ) {
								list.addResolvedConceptReference(subsumed.next());
							}
						}
						else
						{
							rcrl = new ResolvedConceptReferenceList();
							for (Iterator<AssociatedConcept> subsumed = subsumedNodes.iterateAssociatedConcept(); subsumed.hasNext(); ) {
								rcrl.addResolvedConceptReference(subsumed.next());
							}
							hmap.put(assocToCheck, rcrl);
						}
					}
				}
			}

			assocNames = NARROWING_HIER_ASSOC;
			for (int i = 0; i < assocNames.length; i++)
			{
				String assocToCheck = assocNames[i];
				Association subsume = navigateForward
					? getParentsBySource(lbSvc, lbscm, sab, code, null, assocToCheck, scheme, csvt, false)
					: getChildrenBySource(lbSvc, lbscm, sab, code, null, assocToCheck, scheme, csvt, false);

				if (subsume != null) {
					AssociatedConceptList subsumedNodes = subsume.getAssociatedConcepts();
					if (subsumedNodes.getAssociatedConceptCount() > 0)
					{
						if (hmap.get(assocToCheck) != null)
						{
							ResolvedConceptReferenceList list = (ResolvedConceptReferenceList) hmap.get(assocToCheck);
							for (Iterator<AssociatedConcept> subsumed = subsumedNodes.iterateAssociatedConcept(); subsumed.hasNext(); ) {
								list.addResolvedConceptReference(subsumed.next());
							}
						}
						else
						{
							rcrl = new ResolvedConceptReferenceList();
							for (Iterator<AssociatedConcept> subsumed = subsumedNodes.iterateAssociatedConcept(); subsumed.hasNext(); ) {
								rcrl.addResolvedConceptReference(subsumed.next());
							}
							hmap.put(assocToCheck, rcrl);
						}
					}
				}
			}

		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}

		return hmap;
	}


	/**
	 * Return immediate children of the given node within a source SAB
	 * when navigating the specified association.
	 * @param lbscm
	 * @param sab
	 * @param conceptCode
	 * @param relationship
	 * @param association
	 * @param codingScheme
	 * @param versionOrTag
	 * @param buildReferencedEntries
	 * @return
	 * @throws LBException
	 */
	protected static Association getChildrenBySource(
		    LexBIGService lbs,
			LexBIGServiceConvenienceMethods lbscm, String sab,
			String conceptCode, String relationship, String association,
			String codingScheme, CodingSchemeVersionOrTag versionOrTag,
			boolean buildReferencedEntries) throws LBException
	{
		NameAndValueList sourceQualifiers = Constructors.createNameAndValueList(sab, "Source");
		CodedNodeGraph cng = lbs.getNodeGraph(codingScheme, versionOrTag, relationship);
		ConvenienceMethods lbConvMethods = new ConvenienceMethods(lbs);
		cng = cng.restrictToSourceCodes(lbscm.createCodeNodeSet(new String[] { conceptCode }, codingScheme, versionOrTag));
		cng = cng.restrictToAssociations(lbConvMethods.createNameAndValueList(association), sourceQualifiers);

		int buildRefEntries = -1;
		if (buildReferencedEntries) {
			buildRefEntries = 1;
		}
		ResolvedConceptReferenceList rcrl = cng.resolveAsList(ConvenienceMethods.createConceptReference(conceptCode,
				codingScheme), true, false, buildRefEntries, 1, null, null, null, 0);
		if (rcrl.getResolvedConceptReferenceCount() > 0) {
			AssociationList al = rcrl.getResolvedConceptReference(0).getSourceOf();
			if (al.getAssociationCount() > 0) {
				return al.getAssociation(0);
			}
		}
		return null;
	}

	/**
	 * Return immediate parents of the given node within a source SAB
	 * when navigating the given association.
	 * @param lbscm
	 * @param sab
	 * @param conceptCode
	 * @param relationship
	 * @param association
	 * @param codingScheme
	 * @param versionOrTag
	 * @param buildReferencedEntries
	 * @return
	 * @throws LBException
	 */
	protected static Association getParentsBySource(
		    LexBIGService lbs,
			LexBIGServiceConvenienceMethods lbscm, String sab,
			String conceptCode, String relationship, String association,
			String codingScheme, CodingSchemeVersionOrTag versionOrTag,
			boolean buildReferencedEntries) throws LBException
	{
		ConvenienceMethods lbConvMethods = new ConvenienceMethods(lbs);
		NameAndValueList sourceQualifiers = Constructors.createNameAndValueList(sab, "Source");
		CodedNodeGraph cng = lbs.getNodeGraph(codingScheme, versionOrTag, relationship);
		cng = cng.restrictToTargetCodes(lbscm.createCodeNodeSet(new String[] { conceptCode }, codingScheme, versionOrTag));
		cng = cng.restrictToAssociations(lbConvMethods.createNameAndValueList(association), sourceQualifiers);

		int buildRefEntries = -1;
		if (buildReferencedEntries) {
			buildRefEntries = 1;
		}
		ResolvedConceptReferenceList rcrl = cng.resolveAsList(ConvenienceMethods.createConceptReference(conceptCode,
				codingScheme), false, true, buildRefEntries, 1, null, null, null, 0);
		if (rcrl.getResolvedConceptReferenceCount() > 0) {
			AssociationList al = rcrl.getResolvedConceptReference(0).getTargetOf();
			if (al.getAssociationCount() > 0) {
				return al.getAssociation(0);
			}
		}
		return null;
	}

	/**
	 * Retrieves the nodes attached to the system-designated
	 * top node ('@') for the given association and source SAB.
	 * @param codingScheme
	 * @param versionOrTag
	 * @param relationContext
	 * @param association
	 * @param associatedSAB
	 * @return
	 * @throws LBException
	 */

	static boolean isBlank(String s) {
		if (s == null || s.trim().length() == 0) return true;
		return false;
	}

	static boolean isNotBlank(String s) {
		if (s != null && s.trim().length() > 0) return true;
		return false;
	}

	protected static ResolvedConceptReferenceList getTopNodes(
		    LexBIGService lbs,
			String codingScheme,
			CodingSchemeVersionOrTag versionOrTag,
			String relationContext,
			String association,
			String associatedSAB) throws LBException
	{
		CodedNodeGraph cng = lbs.getNodeGraph(codingScheme, versionOrTag,
				relationContext);
		if (isNotBlank(association)) {
			ConvenienceMethods lbConvMethods = new ConvenienceMethods(lbs);
			cng = cng.restrictToAssociations(
				lbConvMethods.createNameAndValueList(association),
				isBlank(associatedSAB) ? null
					: lbConvMethods.createNameAndValueList(associatedSAB, "Source"));
		}

		ResolvedConceptReferenceList list = null;
		try {
			list = cng.resolveAsList(null, true, false, 0, 0, null, null, null, 0);
		} catch (Exception e) {
			System.out.println("Exception thrown - cng.resolveAsList.");
		}
        return list;
	}

	/**
	 * Retrieves the nodes attached to the system-designated
	 * end node ('@@') for the given association and source SAB.
	 * @param codingScheme
	 * @param versionOrTag
	 * @param relationContext
	 * @param association
	 * @param associatedSAB
	 * @return
	 * @throws LBException
	 */
	protected static ResolvedConceptReferenceList getEndNodes(
		    LexBIGService lbs,
			String codingScheme,
			CodingSchemeVersionOrTag versionOrTag,
			String relationContext,
			String association,
			String associatedSAB) throws LBException
	{
		CodedNodeGraph cng = lbs.getNodeGraph(codingScheme, versionOrTag,
				relationContext);
		if (isNotBlank(association)) {
			ConvenienceMethods lbConvMethods = new ConvenienceMethods(lbs);
			cng = cng.restrictToAssociations(
				lbConvMethods.createNameAndValueList(association),
				isBlank(associatedSAB) ? null
					: lbConvMethods.createNameAndValueList(associatedSAB, "Source"));
		}
		ResolvedConceptReferenceList list = null;
		try {
		    list = cng.resolveAsList(null, false, true, 0, 0, null, null, null, 0);
		} catch (Exception e) {
			System.out.println("Exception thrown - cng.resolveAsList.");
		}
        return list;
	}

	private static void displayMessage(String text)
	{
		System.out.println(text);
		pw.println(text);
	}

	private static void displayMessage(PrintWriter pw, String text)
	{
		System.out.println(text);
		pw.println(text);
	}


	public PrintWriter openPrintWriter(String outputfile)
	{
		try {
			PrintWriter pw = new PrintWriter (new BufferedWriter(new FileWriter(outputfile)));
            return pw;
		}catch(Exception e) {
			e.printStackTrace();
			return null;
		}
    }

	public void closeWriter(PrintWriter pw)
	{
		if (pw == null)
		{
			System.out.println("WARNING: closeWriter is not open.");
			return;
		}
		pw.close();
	}
}
